'''cr_lint.py: lint checks'''
#
# Copyright (C) 2013-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from debian.deb822 import Deb822
import glob
import os
import re
import stat
import yaml

from clickreviews.frameworks import Frameworks
from clickreviews.cr_common import (
    ClickReview,
)
from clickreviews.common import (
    open_file_read,
    cmd,
    error,
)
from clickreviews.common import (
    find_external_symlinks,
)

CONTROL_FILE_NAMES = ["control", "manifest", "preinst"]
MINIMUM_CLICK_FRAMEWORK_VERSION = "0.4"


class ClickReviewLint(ClickReview):
    '''This class represents click lint reviews'''

    def __init__(self, fn, overrides=None):
        '''Set up the class.'''
        ClickReview.__init__(self, fn, "lint", overrides=overrides)
        if not self.is_click and not self.is_snap1:
            return

        if self.is_click and "md5sums" not in CONTROL_FILE_NAMES:
            CONTROL_FILE_NAMES.append("md5sums")
        elif self.is_snap1:
            CONTROL_FILE_NAMES.append("hashes.yaml")
        self.control_files = dict()
        self._list_control_files()
        # Valid values for Architecture in DEBIAN/control. Note:
        #  all    - no compiled code
        #  multi  - fat packages (click manifest declares which ones as a list)
        #  armhf  - compiled, single arch (unity8-based devices)
        #  i386   - compiled, single arch (unity8-based emulator and desktop)
        #  amd64  - compiled, single arch (unity8-based desktop)
        #  arm64  - compiled, single arch (not available yet)
        #  ppc64el - compiled, single arch (not available yet)
        #  powerpc - compiled, single arch (not available yet)
        self.valid_compiled_architectures = ['armhf',
                                             'i386',
                                             'amd64',
                                             'arm64',
                                             ]
        self.valid_control_architectures = [
            'all',
            'multi',
        ] + self.valid_compiled_architectures
        self.vcs_files = ['.bzr*',
                          # '.excludes',  # autogenerated by SDK
                          '.git*',
                          '.idea',
                          '.svn*',
                          '.hg',
                          '.project',
                          'CVS*',
                          'RCS*'
                          ]

        if self.manifest is not None and 'maintainer' in self.manifest:
            maintainer = self.manifest['maintainer']
            self.email = maintainer.partition('<')[2].rstrip('>')
            self.is_core_app = \
                (self.click_pkgname.startswith('com.ubuntu.') and
                 not self.click_pkgname.startswith('com.ubuntu.developer.') and
                 (self.email == 'ubuntu-touch-coreapps@lists.launchpad.net' or
                  self.email == 'ubuntu-devel-discuss@lists.ubuntu.com'))
            # "core scope" is not necessarily a word we use right now, but
            # we want to special case scopes which are written through our
            # vetted development process.
            self.is_core_scope = \
                (self.click_pkgname.startswith('com.ubuntu.scopes.') and
                 self.email == 'ubuntu-devel-discuss@lists.ubuntu.com')
            # "core snappy" is not necessarily a word we use right now, but
            # we want to special case scopes which are written through our
            # vetted development process.
            self.is_core_snappy = \
                (self.click_pkgname.startswith('com.ubuntu.snappy.') and
                 self.email == 'ubuntu-devel-discuss@lists.ubuntu.com')
        else:
            self.email = None
            self.is_core_app = False
            self.is_core_scope = False
            self.is_core_snappy = False

        self._list_all_compiled_binaries()

        self.known_hooks = ['accounts',
                            'account-application',
                            'account-provider',
                            'account-qml-plugin',
                            'account-service',
                            'apparmor',
                            'apparmor-profile',
                            'bin-path',  # obsoleted, ignored
                            'content-hub',
                            'desktop',
                            'language-packs',
                            'pay-ui',
                            'puritine',
                            'push-helper',
                            'scope',
                            'snappy-systemd',  # obsoleted, ignored
                            'urls']

        self.redflagged_hooks = ['pay-ui',
                                 'puritine',
                                 'apparmor-profile',
                                 ]

        # Valid values for 'type' in packaging yaml
        # - app
        # - framework
        # - kernel
        # - oem (deprecated, 15.04 only)
        # - gadget
        # - os
        self.snappy_valid_types = ['app',
                                   'framework',
                                   'kernel',
                                   'oem',
                                   'gadget',
                                   'os',
                                   ]
        self.snappy_redflagged_types = ['framework',
                                        'kernel',
                                        'oem',
                                        'gadget',
                                        'os',
                                        ]

    def _list_control_files(self):
        '''List all control files with their full path.'''
        for i in CONTROL_FILE_NAMES:
            self.control_files[i] = os.path.join(self.unpack_dir,
                                                 "DEBIAN/%s" % i)

    def check_control_files(self):
        '''Check DEBIAN/* files'''
        if not self.is_click and not self.is_snap1:
            return

        for f in self.control_files:
            t = 'info'
            n = self._get_check_name(
                'DEBIAN_has_files', extra=os.path.basename(f))
            s = "OK"
            if not os.path.isfile(self.control_files[os.path.basename(f)]):
                if self.is_snap1 and os.path.basename(f) == 'md5sums':
                    s = "OK (skip md5sums with snap)"
                else:
                    t = 'error'
                    s = "'%s' not found in DEBIAN/" % os.path.basename(f)
            self._add_result(t, n, s)

        found = []
        for f in sorted(glob.glob("%s/DEBIAN/*" % self.unpack_dir)):
            if os.path.basename(f) not in self.control_files:
                found.append(os.path.basename(f))
        t = 'info'
        n = self._get_check_name('DEBIAN_extra_files')
        s = 'OK'
        if len(found) > 0:
            t = 'warn'
            s = 'found extra files in DEBIAN/: %s' % ", ".join(found)
        self._add_result(t, n, s)

    def check_control(self):
        '''Check control()'''
        if not self.is_click and not self.is_snap1:
            return

        fh = self._extract_control_file()
        tmp = list(Deb822.iter_paragraphs(fh))
        t = 'info'
        n = self._get_check_name('control_structure')
        s = 'OK'
        if len(tmp) != 1:
            self._add_result('error', n,
                             'control malformed: too many paragraphs')
            return
        self._add_result(t, n, s)

        control = tmp[0]
        fields = ['Package',
                  'Version',
                  'Click-Version',
                  'Architecture',
                  'Maintainer',
                  'Installed-Size',
                  'Description']

        error = False
        for f in sorted(fields):
            t = 'info'
            n = self._get_check_name('control_has_field', extra=f)
            s = 'OK'
            if f not in control:
                if f == 'Maintainer' and self.is_snap1:
                    s = 'OK (maintainer not required for snappy)'
                else:
                    t = 'error'
                    s = "'%s' missing" % f
                    error = True
            self._add_result(t, n, s)
        if error is True:
            return

        t = 'info'
        n = self._get_check_name('control_extra_fields')
        s = 'OK'
        found = []
        for k in sorted(control.keys()):
            if k not in fields:
                found.append(k)
        if len(found) > 0:
            self._add_result('error', n,
                             "found extra fields: '%s'" % (", ".join(found)))

        t = 'info'
        n = self._get_check_name('control_package_match')
        s = "OK"
        if self.manifest['name'] != self.click_pkgname:
            t = 'error'
            s = "Package=%s does not match manifest name=%s" % \
                (self.manifest['name'], self.click_pkgname)
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('control_version_match')
        s = "OK"
        if self.manifest['version'] != self.click_version:
            t = 'error'
            s = "Version=%s does not match manifest version=%s" % \
                (self.manifest['version'], self.click_version)
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('control_architecture_match')
        s = 'OK'
        if 'architecture' in self.manifest:
            if control['Architecture'] == "multi":
                if not isinstance(self.manifest['architecture'],
                                  (list, tuple)):
                    t = 'error'
                    s = 'If arch=multi, manifest architecture needs to ' + \
                        'list all the individual architectures.'
                elif list(filter(lambda a:
                                 a not in self.valid_compiled_architectures,
                                 self.manifest['architecture'])):
                    t = 'error'
                    s = 'If arch=multi, manifest architecture needs to ' + \
                        'comprise of only compiled architectures.'
            elif control['Architecture'] != self.manifest['architecture'] and \
                    not self.is_snap1:  # snappy doesn't use this field; ignore
                t = 'error'
                s = "Architecture=%s " % control['Architecture'] + \
                    "does not match manifest architecture=%s" % \
                    self.manifest['architecture']
        else:  # Lack of architecture in manifest is not an error
            t = 'info'
            s = 'OK: architecture not specified in manifest'
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('control_maintainer_match')
        s = 'OK'
        if 'maintainer' in self.manifest:
            if 'Maintainer' in control and \
                    control['Maintainer'] != self.manifest['maintainer']:
                t = 'error'
                s = "Maintainer=%s does not match manifest maintainer=%s" % \
                    (control['Maintainer'], self.manifest['maintainer'])
        else:
            s = 'Skipped: maintainer not in manifest'
        self._add_result(t, n, s)

        # TODO: click currently sets the Description to be the manifest title.
        # Is this intended behavior?
        t = 'info'
        n = self._get_check_name('control_description_match')
        s = 'OK'
        if 'title' in self.manifest:
            if control['Description'].strip() != \
               self.manifest['title'].strip():
                t = 'error'
                s = "Description=%s does not match manifest title=%s" % \
                    (control['Description'], self.manifest['title'])
        else:
            t = 'warn'
            s = 'Skipped: title not in manifest'
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('control_click_version_up_to_date')
        s = 'OK'
        link = None

        try:
            from apt import apt_pkg
        except Exception:
            s = "SKIPPED (could not import apt_pkg)"
            self._add_result(t, n, s)
            return

        if apt_pkg.version_compare(
                control['Click-Version'], MINIMUM_CLICK_FRAMEWORK_VERSION) < 0:
            t = 'error'
            s = "Click-Version is too old, has '%s', needs '%s' or newer" % (
                control['Click-Version'], MINIMUM_CLICK_FRAMEWORK_VERSION)
            link = ('http://askubuntu.com/questions/417366/what-does-'
                    'lint-control-click-version-up-to-date-mean/417367')
        self._add_result(t, n, s, link)

        t = 'info'
        n = self._get_check_name('control_installed_size')
        s = 'OK'
        try:
            int(control['Installed-Size'])
        except TypeError:
            t = 'error'
            s = "invalid Installed-Size '%s'" % (control['Installed-Size'])
        self._add_result(t, n, s)

    def check_md5sums(self):
        '''Check md5sums()'''
        if not self.is_click:
            return

        t = 'info'
        n = self._get_check_name('md5sums')

        # The puritine click will have bad checksums due to bad symlinks
        if 'hooks' in self.manifest:
            for app in self.manifest['hooks']:
                if "puritine" in self.manifest['hooks'][app]:
                    s = "SKIPPED: puritine"
                    self._add_result(t, n, s)
                    return

        curdir = os.getcwd()
        fh = open_file_read(self.control_files["md5sums"])
        badsums = []
        os.chdir(self.unpack_dir)
        for line in fh.readlines():
            split_line = line.strip().split()
            fn = " ".join(split_line[1:])
            (rc, out) = cmd(['md5sum', fn])
            if line != out:
                badsums.append(fn)
        fh.close()
        os.chdir(curdir)

        s = 'OK'
        if len(badsums) > 0:
            t = 'error'
            s = 'found bad checksums: %s' % ", ".join(badsums)
        self._add_result(t, n, s)

    def check_preinst(self):
        '''Check preinst()'''
        if not self.is_click and not self.is_snap1:
            return

        expected = '''#! /bin/sh
echo "Click packages may not be installed directly using dpkg."
echo "Use 'click install' instead."
exit 1
'''
        fh = open_file_read(self.control_files["preinst"])
        contents = ""
        for line in fh.readlines():
            contents += line
        fh.close()

        t = 'info'
        n = self._get_check_name('preinst')
        s = "OK"
        if contents != expected:
            t = 'error'
            s = "unexpected preinst contents"
        self._add_result(t, n, s)

    def check_hooks(self):
        '''Check click manifest hooks'''
        if not self.is_click and not self.is_snap1:
            return

        #  oem snaps don't have a hooks entry
        if self.is_snap_oem:
            return

        # Some checks are already handled in
        # cr_common.py:_verify_manifest_structure()

        t = 'info'
        n = self._get_check_name('hooks_multiple_apps')
        s = 'OK'
        count = 0
        for app in self.manifest['hooks']:
            if "desktop" in self.manifest['hooks'][app]:
                count += 1
        if count > 1:
            # 'info' for now but this might be removed in a future version
            # (see https://launchpad.net/bugs/1496402)
            t = 'info'
            s = 'more than one desktop app specified in hooks'
        self._add_result(t, n, s)

        # Verify keys are well-formatted
        for app in self.manifest['hooks']:
            t = 'info'
            n = self._get_check_name('hooks_valid', app=app)
            s = "OK"
            if not re.search(r'^[A-Za-z0-9+-.:~-]+$', app):
                t = 'error'
                s = "malformed application name: '%s'" % app
            self._add_result(t, n, s)

        # Verify we have the required hooks
        required = ['apparmor']
        for f in required:
            for app in self.manifest['hooks']:
                t = 'info'
                n = self._get_check_name('hooks', app=app, extra=f)
                s = "OK"
                if f in list(filter(lambda a: a.startswith('account-'),
                   self.known_hooks)):
                    s = "OK (run check-online-accounts for more checks)"
                elif f == "apparmor":
                    s = "OK (run check-security for more checks)"
                elif f == "content-hub":
                    s = "OK (run check-content-hub for more checks)"
                elif f == "desktop":
                    s = "OK (run check-desktop for more checks)"
                elif f == "scope":
                    s = "OK (run check-scope for more checks)"
                elif f == "urls":
                    s = "OK (run check-url-dispatcher for more checks)"

                if f not in self.manifest['hooks'][app]:
                    # TODO: when have apparmor policy for account-provider and
                    #       account-qml-plugin, remove this conditional
                    if f == 'apparmor' and \
                            'apparmor-profile' in self.manifest['hooks'][app]:
                        # Don't require apparmor if have apparmor-profile
                        pass
                    elif f == 'apparmor' and \
                            'puritine' in self.manifest['hooks'][app]:
                        s = "OK (puritine does not use apparmor)"
                    elif f != 'apparmor' or \
                            len(self.manifest['hooks'][app]) != 2 or \
                            'account-provider' not in \
                            self.manifest['hooks'][app] or \
                            'account-qml-plugin' not in \
                            self.manifest['hooks'][app]:
                        t = 'error'
                        s = "'%s' hook not found for '%s'" % (f, app)
                self._add_result(t, n, s)

        mutually_exclusive = ['scope', 'desktop']
        for app in self.manifest['hooks']:
            t = 'info'
            n = self._get_check_name('exclusive_hooks', app=app)
            s = "OK"
            found = []
            for i in mutually_exclusive:
                if i in self.manifest['hooks'][app]:
                    found.append(i)
            if len(found) > 1:
                t = 'error'
                s = "'%s' hooks should not be used together" % ", ".join(found)
            self._add_result(t, n, s)

        for app in self.manifest['hooks']:
            if "apparmor" in self.manifest['hooks'][app]:
                t = 'info'
                n = self._get_check_name('sdk_security_extension', app=app)
                s = "OK"
                fn = self.manifest['hooks'][app]['apparmor']
                if not fn.endswith(".apparmor"):
                    t = 'info'
                    s = ('%s does not end with .apparmor '
                         '(ok if not using sdk)' % fn)
                self._add_result(t, n, s)

    def check_hooks_unknown(self):
        '''Check if have any unknown hooks'''
        if not self.is_click and not self.is_snap1:
            return

        #  oem snaps don't have a hooks entry
        if self.is_snap_oem:
            return

        t = 'info'
        n = self._get_check_name('unknown_hooks')
        s = 'OK'

        # Verify keys are well-formatted
        for app in self.manifest['hooks']:
            for hook in self.manifest['hooks'][app]:
                t = 'info'
                n = self._get_check_name('hooks_known', app=app, extra=hook)
                s = "OK"
                if hook not in self.known_hooks:
                    t = 'warn'
                    s = "unknown hook '%s' in %s" % (hook, app)
                self._add_result(t, n, s)

    def check_hooks_redflagged(self):
        '''Check if have any redflagged hooks'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('redflagged_hooks')
        s = 'OK'

        for app in self.manifest['hooks']:
            found = []
            t = 'info'
            n = self._get_check_name('hooks_redflag', app=app)
            s = "OK"
            manual_review = False
            for hook in self.manifest['hooks'][app]:
                if hook in self.redflagged_hooks:
                    # This check is handled elsewhere
                    if self.is_snap1 and hook == "apparmor-profile":
                        continue
                    found.append(hook)
            if len(found) > 0:
                t = 'error'
                s = "(NEEDS REVIEW) '%s' not allowed" % ", ".join(found)
                manual_review = True
            self._add_result(t, n, s, manual_review=manual_review)

    def check_external_symlinks(self):
        '''Check if symlinks in the click package go out to the system.'''
        if not self.is_click and not self.is_snap1:
            return

        if self.is_snap1 and 'type' in self.pkg_yaml and \
                self.pkg_yaml['type'] not in ['app', 'framework']:
            return

        t = 'info'
        n = self._get_check_name('external_symlinks')
        s = 'OK'

        # The puritine click is expected to have external symlinks
        if 'hooks' in self.manifest:
            for app in self.manifest['hooks']:
                if "puritine" in self.manifest['hooks'][app]:
                    s = "SKIPPED: puritine"
                    self._add_result(t, n, s)
                    return

        links = find_external_symlinks(self.unpack_dir, self.pkg_files,
                                       self.click_pkgname)
        if len(links) > 0:
            t = 'error'
            s = 'package contains external symlinks: %s' % ', '.join(links)
        self._add_result(t, n, s)

    def check_pkgname(self):
        '''Check click package name valid'''
        if not self.is_click:
            return
        p = self.manifest['name']
        # http://www.debian.org/doc/debian-policy/ch-controlfields.html
        t = 'info'
        n = self._get_check_name('pkgname_valid')
        s = "OK"
        if not self._verify_pkgname(p):
            t = 'error'
            s = "'%s' not properly formatted" % p
        self._add_result(t, n, s)

    def check_version(self):
        '''Check click package version is valid'''
        if not self.is_click and not self.is_snap1:
            return

        # deb-version(5)
        t = 'info'
        n = self._get_check_name('version_valid')
        s = "OK"
        # From debian_support.py
        re_valid_version = re.compile(r'^((\d+):)?'              # epoch
                                      '([A-Za-z0-9.+:~-]+?)'     # upstream
                                      '(-([A-Za-z0-9+.~]+))?$')  # debian
        if not re_valid_version.match(self.click_version):
            t = 'error'
            s = "'%s' not properly formatted" % self.click_version
        self._add_result(t, n, s)

    def check_architecture(self):
        '''Check click package architecture in DEBIAN/control is valid'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('control_architecture_valid')
        s = 'OK'
        if self.pkg_arch[0] not in self.valid_control_architectures:
            t = 'error'
            s = "not a valid architecture: %s" % self.pkg_arch[0]
        self._add_result(t, n, s)

    def check_architecture_all(self):
        '''Check if actually architecture all'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('control_architecture_valid_contents')
        s = 'OK'
        if self.pkg_arch[0] != "all":
            self._add_result(t, n, s)
            return

        # look for compiled code
        x_binaries = []
        for i in self.pkg_bin_files:
            x_binaries.append(os.path.relpath(i, self.unpack_dir))
        if len(x_binaries) > 0:
            t = 'error'
            s = "found binaries for architecture 'all': %s" % \
                ", ".join(x_binaries)
        self._add_result(t, n, s)

    def check_architecture_specified_needed(self):
        '''Check if the specified architecture is actually needed'''
        if not self.is_click and not self.is_snap1:
            return

        for arch in self.pkg_arch:
            t = 'info'
            n = self._get_check_name('architecture_specified_needed')
            s = 'OK'
            if arch == "all":
                s = "SKIPPED: architecture is 'all'"
                self._add_result(t, n, s)
                return

            if len(self.pkg_bin_files) == 0:
                t = 'warn'
                s = ("Could not find compiled binaries "
                     "for architecture '%s'" % arch)
            self._add_result(t, n, s)

    def check_maintainer(self):
        '''Check manifest maintainer()'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('maintainer_present')
        s = 'OK'
        if 'maintainer' not in self.manifest:
            if self.is_snap1:
                s = 'Skipped optional maintainer field not specified in ' + \
                    'manifest'
            else:
                t = 'error'
                s = 'required maintainer field not specified in manifest'
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        # Don't perform maintainer checks for snaps. They aren't used by
        # anything.
        if self.is_snap1:
            return

        # Simple regex as used by python3-debian. If we wanted to be more
        # thorough we could use email_re from django.core.validators
        t = 'info'
        n = self._get_check_name('maintainer_format')
        s = 'OK'
        if self.manifest['maintainer'] == "":
            self._add_result('error', n,
                             'invalid maintainer (empty), (should be '
                             'like "Joe Bloggs <joe.bloggs@isp.com>")',
                             'http://askubuntu.com/questions/417351/what-does-'
                             'lint-maintainer-format-mean/417352')
            return
        elif not self._verify_maintainer(self.manifest['maintainer']):
            self._add_result('error', n,
                             'invalid format for maintainer: %s (should be '
                             'like "Joe Bloggs <joe.bloggs@isp.com>")' %
                             self.manifest['maintainer'],
                             'http://askubuntu.com/questions/417351/what-does-'
                             'lint-maintainer-format-mean/417352')
            return
        self._add_result(t, n, s)

    def check_title(self):
        '''Check manifest title()'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('title_present')
        s = 'OK'
        if 'title' not in self.manifest:
            s = 'required title field not specified in manifest'
            self._add_result('error', n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('title')
        s = 'OK'
        pkgname_base = self.click_pkgname.split('.')[-1]
        if len(self.manifest['title']) < len(pkgname_base):
            t = 'info'
            s = "'%s' may be too short" % self.manifest['title']
        self._add_result(t, n, s)

    def check_description(self):
        '''Check manifest description()'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('description_present')
        s = 'OK'
        if 'description' not in self.manifest:
            s = 'required description field not specified in manifest'
            self._add_result('error', n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('description')
        s = 'OK'
        pkgname_base = self.click_pkgname.split('.')[-1]
        if len(self.manifest['description']) < len(pkgname_base):
            t = 'warn'
            if self.is_snap1 and (self.manifest['description'] == '\n' or
                                  self.manifest['description'] == ''):
                s = "manifest description is empty. Is meta/readme.md " + \
                    "formatted correctly?"
            else:
                s = "'%s' is too short" % self.manifest['description']

        self._add_result(t, n, s)

    def check_framework(self):
        '''Check manifest framework()'''
        if not self.is_click and not self.is_snap1:
            return

        # Can check with:
        # click-check-lint <path> '{"framework": {"ubuntu-sdk-15.04": {
        # "state": "obsolete", "policy_vendor": "ubuntu-core",
        # "policy_version": "16.04"}}}'
        n = self._get_check_name('framework')
        link = ("http://askubuntu.com/questions/460512/what-framework-"
                "should-i-use-in-my-manifest-file")
        framework_overrides = self.overrides.get('framework', {})
        frameworks = Frameworks(overrides=framework_overrides)

        if not self.is_snap1 and ',' in self.manifest['framework']:
            # click doesn't support multiple frameworks yet
            t = 'error'
            s = 'ERROR: multiple frameworks found in click manifest'
            self._add_result(t, n, s)
            return

        for framework in self.manifest['framework'].split(','):
            framework = framework.strip()
            if framework in frameworks.OBSOLETE_FRAMEWORKS:
                t = 'error'
                s = "'%s' is obsolete. Please use a newer framework" % \
                    framework
                self._add_result(t, n, s, link)
                return
            elif framework in frameworks.DEPRECATED_FRAMEWORKS:
                t = 'warn'
                s = "'%s' is deprecated. Please use a newer framework" % \
                    framework
                self._add_result(t, n, s, link)
                return
            elif framework in frameworks.AVAILABLE_FRAMEWORKS:
                t = 'info'
                s = 'OK'
                self._add_result(t, n, s)
                return
            else:
                # None of the above checks triggered, this is an unknown
                # framework
                t = 'error'
                s = "'%s' is not a supported framework" % \
                    framework
                self._add_result(t, n, s, link)

    def check_click_local_extensions(self):
        '''Report any click local extensions'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('click_local_extensions')
        s = 'OK'
        found = []
        for k in sorted(self.manifest):
            if k.startswith('x-'):
                found.append(k)

        # Some local extensions are ok for core apps, so just skip them
        if self.is_core_app:
            for i in ['x-source', 'x-test']:
                if i in found:
                    found.remove(i)

        if len(found) > 0:
            t = 'warn'
            plural = ""
            if len(found) > 1:
                plural = "s"
            s = 'found unofficial extension%s: %s' % (plural, ', '.join(found))
        self._add_result(t, n, s)

    def check_vcs(self):
        '''Check for VCS files in the package'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('vcs_files')
        s = 'OK'
        found = []
        for d in self.vcs_files:
            entries = glob.glob("%s/%s" % (self.unpack_dir, d))
            if len(entries) > 0:
                for i in entries:
                    found.append(os.path.relpath(i, self.unpack_dir))
        if len(found) > 0:
            t = 'warn'
            s = 'found VCS files in package: %s' % ", ".join(found)
        self._add_result(t, n, s)

    def check_click_in_package(self):
        '''Check for *.click files in the toplevel click package'''
        if not self.is_click:
            return

        t = 'info'
        n = self._get_check_name('click_files')
        s = 'OK'
        found = []
        entries = glob.glob("%s/*.click" % self.unpack_dir)
        if len(entries) > 0:
            for i in entries:
                found.append(os.path.relpath(i, self.unpack_dir))
        if len(found) > 0:
            t = 'warn'
            s = 'found click packages in toplevel dir: %s' % ", ".join(found)
        self._add_result(t, n, s)

    def check_dot_click(self):
        '''Check for .click directory in the toplevel click package'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('dot_click')
        s = 'OK'

        path = os.path.join(self.unpack_dir, '.click')
        if os.path.exists(path):
            t = 'error'
            s = 'found .click in toplevel dir'
        self._add_result(t, n, s)

    def check_contents_for_hardcoded_paths(self):
        '''Check for known hardcoded paths.'''
        if not self.is_click and not self.is_snap1:
            return

        PATH_BLACKLIST = ["/opt/click.ubuntu.com/"]
        t = 'info'
        n = self._get_check_name('hardcoded_paths')
        s = 'OK'
        for dirpath, dirnames, filenames in os.walk(self.unpack_dir):
            for filename in filenames:
                full_fn = os.path.join(dirpath, filename)
                (rc, out) = cmd(['file', '-b', full_fn])
                if 'text' not in out:
                    continue
                try:
                    lines = open_file_read(full_fn).readlines()
                    for bad_path in PATH_BLACKLIST:
                        if list(filter(lambda line: bad_path in line, lines)):
                            t = 'error'
                            s = "Hardcoded path '%s' found in '%s'." % (
                                bad_path, full_fn)
                except FileNotFoundError:
                    pass
                except UnicodeDecodeError:
                    pass
        self._add_result(t, n, s)

    def _verify_architecture(self, my_dict, test_str):
        t = 'info'
        n = self._get_check_name('%s_architecture_valid' % test_str)
        s = 'OK'
        if 'architecture' not in my_dict and 'architectures' not in my_dict:
            s = 'OK (architecture not specified)'
            self._add_result(t, n, s)
            return

        key = 'architecture'
        if self.is_snap1 and 'architectures' in my_dict:
            # new yaml allows for 'architecture' and 'architectures'
            key = 'architectures'

        archs_list = list(self.valid_control_architectures)
        archs_list.remove("multi")

        if self.is_snap1 and key == 'architectures' and \
           isinstance(my_dict[key], str):
            # new yaml uses 'architectures' that must be a list
            t = 'error'
            s = "not a valid architecture: %s (not a list)" % my_dict[key]
        elif isinstance(my_dict[key], str) and my_dict[key] not in archs_list:
            t = 'error'
            s = "not a valid architecture: %s" % my_dict[key]
        elif isinstance(my_dict[key], list):
            if not self.is_snap1:
                archs_list.remove("all")
            bad_archs = []
            for a in my_dict[key]:
                if a not in archs_list:
                    bad_archs.append(a)
                if len(bad_archs) > 0:
                    t = 'error'
                    s = "not valid multi architecture: %s" % \
                        ",".join(bad_archs)
        self._add_result(t, n, s)

    def check_manifest_architecture(self):
        '''Check package architecture in manifest is valid'''
        if not self.is_click and not self.is_snap1:
            return

        self._verify_architecture(self.manifest, "manifest")

    def _verify_icon(self, my_dict, test_str):
        t = 'info'
        n = self._get_check_name('%s_icon_present' % test_str)
        s = 'OK'
        if 'icon' not in my_dict:
            s = 'Skipped, optional icon not present'
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('%s_icon_empty' % test_str)
        s = 'OK'
        if len(my_dict['icon']) == 0:
            t = 'error'
            s = "icon entry is empty"
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('%s_icon_absolute_path' % test_str)
        s = 'OK'
        if my_dict['icon'].startswith('/'):
            t = 'error'
            s = "icon entry '%s' should not specify absolute path" % \
                my_dict['icon']
        self._add_result(t, n, s)

    def check_icon(self):
        '''Check icon()'''
        if not self.is_click and not self.is_snap1:
            return

        self._verify_icon(self.manifest, "manifest")

    def check_snappy_name(self):
        '''Check package name'''
        if self.is_click or not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('snappy_name_valid')
        s = 'OK'
        if 'name' not in self.pkg_yaml:
            t = 'error'
            s = "could not find 'name' in yaml"
        elif not self._verify_pkgname(self.pkg_yaml['name']):
            t = 'error'
            s = "malformed 'name': '%s'" % self.pkg_yaml['name']
        self._add_result(t, n, s)

    def check_snappy_version(self):
        '''Check package version'''
        if self.is_click or not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('snappy_version_valid')
        s = 'OK'
        if 'version' not in self.pkg_yaml:
            t = 'error'
            s = "could not find 'version' in yaml"
        elif not self._verify_pkgversion(self.pkg_yaml['version']):
            t = 'error'
            s = "malformed 'version': '%s'" % self.pkg_yaml['version']
        self._add_result(t, n, s)

    def check_snappy_type(self):
        '''Check type'''
        if self.is_click or not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('snappy_type_valid')
        s = 'OK'
        if 'type' not in self.pkg_yaml:
            s = 'OK (skip missing)'
        elif self.pkg_yaml['type'] not in self.snappy_valid_types:
            t = 'error'
            s = "unknown 'type': '%s'" % self.pkg_yaml['type']
        self._add_result(t, n, s)

    def check_snappy_type_redflagged(self):
        '''Check if snappy type is redflagged'''
        if self.is_click or not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('snappy_type_redflag')
        s = "OK"
        link = None
        manual_review = False
        if 'type' not in self.pkg_yaml:
            s = 'OK (skip missing)'
        elif self.pkg_yaml['type'] in self.snappy_redflagged_types:
            t = 'error'
            s = "(NEEDS REVIEW) type '%s' not allowed" % self.pkg_yaml['type']
            manual_review = True
            if self.pkg_yaml['type'] == "framework":
                link = ("https://developer.ubuntu.com/en/snappy/"
                        "guides/frameworks/")
        self._add_result(t, n, s, link=link, manual_review=manual_review)

    def check_snappy_icon(self):
        '''Check icon()'''
        if self.is_click or not self.is_snap1:
            return

        self._verify_icon(self.pkg_yaml, "package_yaml")

    def check_snappy_architecture(self):
        '''Check package architecture in package.yaml is valid'''
        if self.is_click or not self.is_snap1:
            return

        self._verify_architecture(self.pkg_yaml, "package yaml")

        t = 'info'
        n = self._get_check_name('snappy_architecture_deprecated')
        s = 'OK'
        link = None
        if 'architecture' in self.pkg_yaml:
            t = 'warn'
            s = "Found deprecated 'architecture' field in " + \
                "meta/package.yaml. Use 'architectures' instead"
            link = ('https://developer.ubuntu.com/en/snappy/'
                    'guides/package-metadata/')
        self._add_result(t, n, s, link=link)

    def check_snappy_unknown_entries(self):
        '''Check for any unknown fields'''
        if self.is_click or not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('snappy_unknown')
        s = 'OK'
        unknown = []
        for f in self.pkg_yaml:
            if f not in self.snappy_required + self.snappy_optional:
                unknown.append(f)
        if len(unknown) > 0:
            t = 'warn'
            s = "unknown entries in package.yaml: '%s'" % \
                (",".join(sorted(unknown)))
            obsoleted = ['maintainer', 'ports']
            tmp = list(set(unknown) & set(obsoleted))
            if len(tmp) > 0:
                t = 'error'
                s += " (%s obsoleted)" % ",".join(tmp)
        self._add_result(t, n, s)

    def _extract_readme_md(self):
        '''Extract meta/readme.md'''
        contents = None
        readme = os.path.join(self.unpack_dir, "meta/readme.md")
        if os.path.exists(readme):
            fh = open_file_read(readme)
            contents = fh.read()
        return contents

    def check_snappy_readme_md(self):
        '''Check snappy readme.md'''
        if self.is_click or not self.is_snap1:
            return

        contents = self._extract_readme_md()

        t = 'info'
        n = self._get_check_name('snappy_readme.md')
        s = 'OK'
        if contents is None:
            t = 'error'
            s = 'meta/readme.md does not exist'
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('snappy_readme.md_length')
        s = 'OK'
        pkgname_base = self.pkg_yaml['name'].split('.')[0]
        if len(contents) < len(pkgname_base):
            t = 'warn'
            s = "meta/readme.md is too short"
        self._add_result(t, n, s)

    def check_snappy_config(self):
        '''Check snappy config'''
        if self.is_click or not self.is_snap1:
            return

        fn = os.path.join(self.unpack_dir, 'meta/hooks/config')
        if fn not in self.pkg_files:
            return

        t = 'info'
        n = self._get_check_name('snappy_config_hook_executable')
        s = 'OK'
        if not self._check_innerpath_executable(fn):
            t = 'error'
            s = 'meta/hooks/config is not executable'
        self._add_result(t, n, s)

    def check_snappy_services_and_binaries(self):
        '''Services and binaries should not overlap'''
        if self.is_click or not self.is_snap1:
            return
        for exe_t in ['binaries', 'services']:
            if exe_t not in self.pkg_yaml:
                break
            if exe_t == 'binaries':
                other_exe_t = 'services'
            else:
                other_exe_t = 'binaries'

            if other_exe_t not in self.pkg_yaml:
                break

            for a in self.pkg_yaml[exe_t]:
                if 'name' not in a:
                    continue
                # handle 'exec' in binaries
                app = os.path.basename(a['name'])

                t = 'info'
                n = self._get_check_name(
                    'snappy_in_%s' % other_exe_t, app=app)
                s = 'OK'
                for other_a in self.pkg_yaml[other_exe_t]:
                    other_app = os.path.basename(other_a['name'])
                    if app == other_app:
                        t = 'error'
                        s = "'%s' in both 'services' and 'binaries'" % app
                        break
                self._add_result(t, n, s)

    def check_snappy_hashes(self):
        '''Check snappy hashes.yaml'''
        if self.is_click or not self.is_snap1:
            return

        def _check_allowed_perms(mode, allowed):
            '''Check that mode only uses allowed perms'''
            for p in mode[1:]:
                if p not in allowed:
                    return False
            return True

        try:
            hashes_yaml = yaml.safe_load(self._extract_hashes_yaml())
        except Exception:
            error("Could not load hashes.yaml. Is it properly formatted?")

        if 'archive-sha512' not in hashes_yaml:
            t = 'error'
            n = self._get_check_name('hashes_archive-sha512_present')
            s = "'archive-sha512' not found in hashes.yaml"
            self._add_result(t, n, s)
            return

        # verify the ar file
        t = 'info'
        n = self._get_check_name('hashes_archive-sha512_valid')
        s = 'OK'
        fn = self._path_join(self.raw_unpack_dir, 'data.tar.gz')
        sum = self._get_sha512sum(fn)
        if hashes_yaml['archive-sha512'] != sum:
            t = 'error'
            s = "hash mismatch: '%s' != '%s'" % (hashes_yaml['archive-sha512'],
                                                 sum)
            self._add_result(t, n, s)
            return
        self._add_result(t, n, s)

        if 'files' not in hashes_yaml:
            t = 'error'
            n = self._get_check_name('hashes_files_present')
            s = "'files' not found in hashes.yaml"
            self._add_result(t, n, s)
            return

        # verify the individual files
        errors = []
        badsums = []
        hash_files = set([])  # used to check with extra files
        for entry in hashes_yaml['files']:
            if 'name' not in entry:
                errors.append("'name' not found for entry '%s'" % entry)
                continue
            else:
                hash_files.add(entry['name'])

            if 'mode' not in entry:
                errors.append("'mode' not found for entry '%s'" %
                              entry['name'])
                continue
            elif len(entry['mode']) != 10:
                errors.append("malformed mode '%s' for entry '%s'" %
                              (entry['mode'], entry['name']))
                continue
            elif entry['mode'].startswith('d'):
                if not _check_allowed_perms(entry['mode'],
                                            ['r', 'w', 'x', '-']):
                    errors.append("unusual mode '%s' for entry '%s'" %
                                  (entry['mode'], entry['name']))
                # world write shouldn't be allowed
                if entry['mode'][-2] != '-':
                    errors.append("'%s' is world-writable" % entry['name'])
                continue
            elif entry['mode'].startswith('l'):
                if entry['mode'] != "lrwxrwxrwx":
                    errors.append("unusual mode '%s' for entry '%s'" %
                                  (entry['mode'], entry['name']))
                continue
            elif not entry['mode'].startswith('f'):
                # files and symlinks are ok, everything else is not
                errors.append("illegal file mode '%s': '%s' for '%s'" %
                              (entry['mode'][0], entry['mode'], entry['name']))
                continue
            elif 'size' not in entry:
                errors.append("'size' not found for entry: %s" % entry['name'])
                continue
            elif 'sha512' not in entry:
                errors.append("'sha512' not found for entry: %s" %
                              entry['name'])
                continue

            fn = self._path_join(self.unpack_dir, entry['name'])

            # quick verify the size, if it is wrong, we don't have to do the
            # sha512sum
            statinfo = self._extract_statinfo(fn)
            # file is missing
            if statinfo is None:
                errors.append("'%s' does not exist" % entry['name'])
                continue
            if entry['size'] != statinfo.st_size:
                errors.append("size %d != %d for '%s'" % (entry['size'],
                                                          statinfo.st_size,
                                                          entry['name']))
                continue

            # check for weird perms
            if not _check_allowed_perms(entry['mode'], ['r', 'w', 'x', '-']):
                errors.append("unusual mode '%s' for entry '%s'" %
                              (entry['mode'], entry['name']))
                continue
            # world write shouldn't be allowed
            if entry['mode'][-2] != '-':
                errors.append("mode '%s' for '%s' is world-writable" %
                              (entry['mode'], entry['name']))
                continue
            # verify perms match what is in hashes.yaml
            filemode = stat.filemode(statinfo.st_mode)[1:]
            if entry['mode'][1:] != filemode:
                errors.append("mode '%s' != '%s' for '%s'" %
                              (entry['mode'][1:], filemode, entry['name']))
                continue

            # ok, now all the cheap tests are done so we can check if we have a
            # valid sha512sum
            sum = self._get_sha512sum(fn)
            if entry['sha512'] != sum:
                badsums.append("'%s' != '%s' for '%s'" % (entry['sha512'], sum,
                                                          entry['name']))

        t = 'info'
        n = self._get_check_name('sha512sums')
        s = 'OK'
        if len(badsums) > 0:
            t = 'error'
            s = 'found bad checksums: %s' % ", ".join(badsums)
        self._add_result(t, n, s)

        t = 'info'
        n = self._get_check_name('file_mode')
        s = 'OK'
        if len(errors) > 0:
            t = 'error'
            s = 'found errors in hashes.yaml: %s' % ", ".join(errors)
        self._add_result(t, n, s)

        # Now check for extra files
        t = 'info'
        n = self._get_check_name('hashes_extra_files')
        s = 'OK'
        self._add_result(t, n, s)
        extra = []
        click_compat_files = set(["DEBIAN/hashes.yaml",
                                  "DEBIAN/control",
                                  "DEBIAN/preinst",
                                  "DEBIAN/manifest"
                                  ])
        for f in self.pkg_files:
            fn = os.path.relpath(f, self.unpack_dir)
            if fn not in hash_files and fn not in click_compat_files:
                extra.append(fn)

        if len(extra) > 0:
            t = 'error'
            s = 'found extra files not listed in hashes.yaml: %s' % \
                ", ".join(extra)
        self._add_result(t, n, s)

    def check_snappy_frameworks(self):
        '''TODO'''
        return False



#
#  Copyright (C) 2014-2016 Canonical Ltd.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import re
from socket import timeout
import sys
import time
from urllib import request, parse
from urllib.error import HTTPError, URLError
import yaml

DATA_DIR = os.path.join(os.path.expanduser('~/.cache/click-reviewers-tools/'))
UPDATE_INTERVAL = 60 * 60 * 24 * 7


def _update_is_necessary(fn):
    return (not os.path.exists(fn)) or \
        (time.time() - os.path.getmtime(fn) >= UPDATE_INTERVAL)


def _update_is_possible(url):
    update = True
    try:
        request.urlopen(url)
    except (HTTPError, URLError):
        update = False
    except timeout:
        update = False
    return update


def abort(msg=None):
    if msg:
        print(msg, file=sys.stderr)
    print('Aborted.', file=sys.stderr)
    sys.exit(1)


#
# Public
#
def get_remote_data(url):
    try:
        f = request.urlopen(url)
    except (HTTPError, URLError) as error:
        abort('Data not retrieved because %s.' % error)
    except timeout:
        abort('Socket timed out.')
    if not f:
        abort()
    return f.read()


def get_remote_file_url(url):
    html = get_remote_data(url)
    # XXX: This is a hack and will be gone, as soon as myapps has an API for
    # this.
    link = re.findall(r'<a href="(\S+?)">download file</a>', html)
    if not link:
        abort()
    download_link = '{}://{}/{}'.format(
        parse.urlparse(url).scheme,
        parse.urlparse(url).netloc,
        link[0].decode("utf-8"))
    return download_link


def get_remote_file(fn, url, data_dir=DATA_DIR):
    data = get_remote_data(url)
    if os.path.exists(fn):
        os.remove(fn)
    if not os.path.exists(os.path.dirname(fn)):
        os.makedirs(os.path.dirname(fn))
    with open(fn, 'bw') as local_file:
        local_file.write(data)


def read_cr_file(fn, url, local_copy_fn=None, as_yaml=False):
    '''read click reviews file from remote or local copy:
       - fn: where to store the cached file
       - url: url to fetch
       - local_copy_fn: force use of local copy
    '''
    d = {}
    if local_copy_fn and os.path.exists(local_copy_fn):
        try:
            if as_yaml:
                d = yaml.safe_load(open(local_copy_fn, 'r').read())
            else:
                d = json.loads(open(local_copy_fn, 'r').read())
        except ValueError:
            raise ValueError("Could not parse '%s'" % local_copy_fn)
    else:
        if _update_is_necessary(fn) and _update_is_possible(url):
            get_remote_file(fn, url)
        if os.path.exists(fn):
            try:
                if as_yaml:
                    d = yaml.safe_load(open(fn, 'r').read())
                else:
                    d = json.loads(open(fn, 'r').read())
            except ValueError:
                raise ValueError("Could not parse '%s'" % fn)
    return d



'''cr_bin_path.py: click bin-path'''
#
# Copyright (C) 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from clickreviews.cr_common import ClickReview, error
import os


class ClickReviewBinPath(ClickReview):
    '''This class represents click lint reviews'''
    def __init__(self, fn, overrides=None):
        # bin-path is ignored by snappy install so don't bother with peerhooks
        ClickReview.__init__(self, fn, "bin-path", overrides=overrides)

        self.bin_paths_files = dict()
        self.bin_paths = dict()

        if not self.is_snap1:
            return

        # snappy yaml currently only allows specifying:
        # - exec (optional)
        # - description (optional)
        self.required_keys = []
        self.optional_keys = ['description', 'exec'] + self.snappy_exe_security

        if self.is_snap1 and 'binaries' in self.pkg_yaml:
            if len(self.pkg_yaml['binaries']) == 0:
                error("package.yaml malformed: 'binaries' is empty")
            for binary in self.pkg_yaml['binaries']:
                if 'name' not in binary:
                    error("package.yaml malformed: required 'name' not found "
                          "for entry in %s" % self.pkg_yaml['binaries'])
                elif not isinstance(binary['name'], str):
                    error("package.yaml malformed: required 'name' is not str"
                          "for entry in %s" % self.pkg_yaml['binaries'])

                app = os.path.basename(binary['name'])
                if 'exec' in binary:
                    rel = binary['exec']
                else:
                    rel = binary['name']
                self.bin_paths[app] = rel
                self.bin_paths_files[app] = self._extract_bin_path(app)

    def _extract_bin_path(self, app):
        '''Get bin-path for app'''
        rel = self.bin_paths[app]
        fn = os.path.join(self.unpack_dir, rel)
        if not os.path.exists(fn):
            error("Could not find '%s'" % rel)
        return fn

    def _check_bin_path_executable(self, app):
        '''Check that the provided path exists'''
        fn = self.bin_paths_files[app]
        return os.access(fn, os.X_OK)

    def _verify_required(self, my_dict, test_str):
        for app in sorted(my_dict):
            for r in self.required_keys:
                found = False
                t = 'info'
                n = self._get_check_name(
                    '%s_required_key' % test_str, app=app, extra=r)
                s = "OK"
                if r in my_dict[app]:
                    if not isinstance(my_dict[app][r], str):
                        t = 'error'
                        s = "'%s' is not a string" % r
                    elif my_dict[app][r] == "":
                        t = 'error'
                        s = "'%s' is empty" % r
                    else:
                        found = True
                if not found and t != 'error':
                    t = 'error'
                    s = "Missing required field '%s'" % r
                self._add_result(t, n, s)

    def check_snappy_required(self):
        '''Check for package.yaml required fields'''
        if not self.is_snap1 or 'binaries' not in self.pkg_yaml:
            return
        self._verify_required(self._create_dict(self.pkg_yaml['binaries']),
                              'package_yaml')

    def _verify_optional(self, my_dict, test_str):
        for app in sorted(my_dict):
            for o in self.optional_keys:
                if o in self.snappy_exe_security:
                    continue  # checked in cr_security.py
                found = False
                t = 'info'
                n = self._get_check_name(
                    '%s_optional_key' % test_str, app=app, extra=o)
                s = "OK"
                if o in my_dict[app]:
                    if not isinstance(my_dict[app][o], str):
                        t = 'error'
                        s = "'%s' is not a string" % o
                    elif my_dict[app][o] == "":
                        t = 'error'
                        s = "'%s' is empty" % o
                    else:
                        found = True
                if not found and t != 'error':
                    s = "OK (skip missing)"
                self._add_result(t, n, s)

    def check_snappy_optional(self):
        '''Check snappy packate.yaml optional fields'''
        if not self.is_snap1 or 'binaries' not in self.pkg_yaml:
            return
        self._verify_optional(self._create_dict(self.pkg_yaml['binaries']),
                              'package_yaml')

    def _verify_unknown(self, my_dict, test_str):
        for app in sorted(my_dict):
            unknown = []
            t = 'info'
            n = self._get_check_name(
                '%s_unknown_key' % test_str, app=app)
            s = "OK"

            for f in my_dict[app].keys():
                if f not in self.required_keys and \
                   f not in self.optional_keys:
                    unknown.append(f)

            if len(unknown) == 1:
                t = 'warn'
                s = "Unknown field '%s'" % unknown[0]
            elif len(unknown) > 1:
                t = 'warn'
                s = "Unknown fields '%s'" % ", ".join(unknown)
            self._add_result(t, n, s)

    def check_snappy_unknown(self):
        '''Check snappy package.yaml unknown fields'''
        if not self.is_snap1 or 'binaries' not in self.pkg_yaml:
            return
        self._verify_unknown(self._create_dict(self.pkg_yaml['binaries']),
                             'package_yaml')

    def check_path(self):
        '''Check path exists'''
        if not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('path_exists')
        s = "OK"

        for app in sorted(self.bin_paths_files):
            t = 'info'
            n = self._get_check_name('path_executable')
            s = "OK"
            if not self._check_bin_path_executable(app):
                t = 'error'
                s = "'%s' is not executable" % \
                    os.path.relpath(self.bin_paths_files[app], self.unpack_dir)
            self._add_result(t, n, s)

    def check_binary_description(self):
        '''Check package.yaml binary description'''
        if not self.is_snap1 or 'binaries' not in self.pkg_yaml:
            return

        my_dict = self._create_dict(self.pkg_yaml['binaries'])

        for app in sorted(my_dict):
            t = 'info'
            n = self._get_check_name(
                'package_yaml_description_present', app=app)
            s = 'OK'
            if 'description' not in my_dict[app]:
                s = 'OK (skip missing)'
                self._add_result('info', n, s)
                return
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name(
                'package_yaml_description_empty', app=app)
            s = 'OK'
            if len(my_dict[app]['description']) == 0:
                t = 'error'
                s = "description is empty"
            self._add_result(t, n, s)



'''cr_security.py: click security checks'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from clickreviews.common import (
    AA_PROFILE_NAME_MAXLEN,
    AA_PROFILE_NAME_ADVLEN,
)
from clickreviews.cr_common import ClickReview, error, open_file_read
import clickreviews.cr_common as cr_common
from clickreviews.apparmor import APPARMOR_POLICIES
import copy
import json
import os


class ClickReviewSecurity(ClickReview):
    '''This class represents click lint reviews'''
    def __init__(self, fn, overrides=None):
        peer_hooks = dict()
        my_hook = 'apparmor'
        peer_hooks[my_hook] = dict()
        # Basically, everything except frameworks
        peer_hooks[my_hook]['allowed'] = ClickReview.app_allowed_peer_hooks + \
            ClickReview.scope_allowed_peer_hooks + \
            ClickReview.service_allowed_peer_hooks + \
            ['pay-ui', 'language-packs']
        peer_hooks[my_hook]['required'] = []

        my_hook2 = 'apparmor-profile'
        peer_hooks[my_hook2] = dict()
        # Basically, everything except frameworks
        peer_hooks[my_hook2]['allowed'] = \
            ClickReview.service_allowed_peer_hooks
        peer_hooks[my_hook2]['required'] = []

        ClickReview.__init__(self, fn, "security", peer_hooks=peer_hooks,
                             overrides=overrides)

        if not self.is_click and not self.is_snap1:
            return

        self.aa_policy = APPARMOR_POLICIES

        self.all_fields = ['abstractions',
                           'author',
                           'binary',
                           'comment',
                           'copyright',
                           'name',
                           'policy_groups',
                           'policy_vendor',
                           'policy_version',
                           'read_path',
                           'template',
                           'template_variables',
                           'write_path']
        self.ignored_fields = ['author',
                               'comment',
                               'copyright',
                               'name']
        self.required_fields = ['policy_version']
        self.redflag_fields = ['abstractions',
                               'binary',
                               'read_path',
                               'template_variables',
                               'write_path']
        self.allowed_webapp_policy_groups = ['accounts',
                                             'audio',
                                             'camera',
                                             'connectivity',
                                             'content_exchange',
                                             'content_exchange_source',
                                             'keep-display-on',
                                             'location',
                                             'microphone',
                                             'networking',
                                             'video',
                                             'webview']

        self.allowed_push_helper_policy_groups = ['push-notification-client']
        self.allowed_network_scope_policy_groups = ['accounts', 'networking']
        self.required_ubuntu_account_plugin_policy_groups = ['accounts',
                                                             'networking']
        self.allowed_ubuntu_account_plugin_policy_groups = \
            self.required_ubuntu_account_plugin_policy_groups + \
            ['audio', 'webview']

        self.redflag_templates = ['unconfined']
        # TODO: how to deal with other vendors
        self.extraneous_ubuntu_templates = ['ubuntu-sdk',
                                            'default']

        # framework policy is based on major framework version. In 13.10, there
        # was only 'ubuntu-sdk-13.10', but in 14.04, there will be several,
        # like 'ubuntu-sdk-14.04-html5', 'ubuntu-sdk-14.04-platform', etc
        # TODO: in 15.10 we will be using the policy version will match the
        # release framework, so we can simplify these checks
        self.major_framework_policy = {
            'ubuntu-sdk-13.10': {
                'policy_version': 1.0,
            },
            'ubuntu-sdk-14.04': {
                'policy_version': 1.1,
            },
            'ubuntu-sdk-14.10': {
                'policy_version': 1.2,
            },
            'ubuntu-sdk-15.04': {
                'policy_version': 1.3,
            },
            'ubuntu-sdk-16.04': {
                'policy_version': 16.04,
            },
            'ubuntu-sdk-20.04': {
                'policy_version': 20.04,
            },
            'ubuntu-core-15.04': {
                'policy_vendor': 'ubuntu-core',
                'policy_version': 15.04,
            }
        }
        framework_overrides = self.overrides.get('framework', {})
        self._override_framework_policies(framework_overrides)

        # snappy
        self.sec_skipped_types = ['oem',
                                  'os',
                                  'kernel']  # these don't need security items

        self.security_manifests = dict()
        self.security_apps = []
        self.security_profiles = dict()
        self.security_apps_profiles = []

        if self.manifest is None and self.is_snap1:
            for exe_t in ['services', 'binaries']:
                if exe_t not in self.pkg_yaml:
                    continue
                for item in self.pkg_yaml[exe_t]:
                    if 'name' not in item:
                        continue
                    app = "%s/%s" % (exe_t, item['name'])

                    if 'security-policy' in item:
                        if 'apparmor' not in item['security-policy']:
                            continue
                        rel_fn = item['security-policy']['apparmor']
                        self.security_profiles[rel_fn] = \
                            self._extract_security_profile(app)
                        self.security_apps_profiles.append(app)
                        continue

                    # Fake a security manifest for code reuse
                    # FIXME: this needs to be updated when we have 'target'
                    m = dict()
                    m['policy_vendor'] = "ubuntu-core"
                    m['policy_version'] = self._pkgfmt_version()
                    if 'security-template' in item:
                        m['template'] = item['security-template']
                    else:
                        m['template'] = 'default'

                    if 'caps' in item:
                        m['policy_groups'] = item['caps']
                    elif self._pkgfmt_version() == "15.04":
                        m['policy_groups'] = ['networking']
                    else:
                        m['policy_groups'] = ['network-client']

                    self.security_manifests[app] = m
                    self.security_apps.append(app)
        else:
            self.security_manifests = dict()
            self.security_apps = []
            hooks = self.manifest['hooks']
            for app in hooks:
                if 'apparmor' not in hooks[app]:
                    #  msg("Skipped missing apparmor hook for '%s'" % app)
                    continue
                if not isinstance(hooks[app]['apparmor'], str):
                    error("manifest malformed: hooks/%s/apparmor "
                          "is not str" % app)
                rel_fn = hooks[app]['apparmor']
                self.security_manifests[rel_fn] = \
                    self._extract_security_manifest(app)
                self.security_apps.append(app)

            for app in hooks:
                if 'apparmor-profile' not in hooks[app]:
                    #  msg("Skipped missing apparmor hook for '%s'" % app)
                    continue
                if not isinstance(hooks[app]['apparmor-profile'], str):
                    error("manifest malformed: hooks/%s/apparmor-profile "
                          "is not str" % app)
                rel_fn = hooks[app]['apparmor-profile']
                self.security_profiles[rel_fn] = \
                    self._extract_security_profile(app)
                self.security_apps_profiles.append(app)

    def _override_framework_policies(self, overrides):
        # override major framework policies
        self.major_framework_policy.update(overrides)

        # override apparmor policies
        for name, data in overrides.items():
            vendor = data.get('policy_vendor')
            version = str(data.get('policy_version'))

            if vendor not in self.aa_policy:
                self.aa_policy[vendor] = {}

            if version not in self.aa_policy[vendor]:
                # just ensure the version is defined
                # TODO: add support to override templates and policy groups
                self.aa_policy[vendor][version] = {}

    def _extract_security_manifest(self, app):
        '''Extract security manifest and verify it has the expected
           structure'''
        d = self.manifest['hooks'][app]['apparmor']
        fn = os.path.join(self.unpack_dir, d)
        rel_fn = self.manifest['hooks'][app]['apparmor']

        try:
            m = json.load(cr_common.open_file_read(fn))
        except Exception:
            error("Could not load '%s'. Is it properly formatted?" % rel_fn)
        mp = json.dumps(m, sort_keys=True, indent=2, separators=(',', ': '))
        if not isinstance(m, dict):
            error("'%s' malformed:\n%s" % (rel_fn, mp))
        for k in sorted(m):
            if k not in self.all_fields:
                error("'%s' malformed: unsupported field '%s':\n%s" % (rel_fn,
                                                                       k, mp))
            if k in ['abstractions', 'policy_groups', 'read_path',
                     'write_path']:
                if not isinstance(m[k], list):
                    error("'%s' malformed: '%s' is not list:\n%s" % (rel_fn,
                                                                     k, mp))
            elif k == 'template_variables':
                if not isinstance(m[k], dict):
                    error("'%s' malformed: '%s' is not dict:\n%s" % (rel_fn,
                                                                     k, mp))
            elif k == "policy_version":
                # python and Qt don't agree on the JSON output of floats that
                # are integers (ie, 1.0 vs 1). LP: #1214618
                if not isinstance(m[k], float) and not isinstance(m[k], int):
                    error("'%s' malformed: '%s' is not a JSON number:\n%s" %
                          (rel_fn, k, mp))
                if isinstance(m[k], int):
                    m[k] = float(m[k])
            else:
                if not isinstance(m[k], str):
                    error("'%s' malformed: '%s' is not str:\n%s" % (rel_fn,
                                                                    k, mp))
        return m

    def _get_security_manifest(self, app):
        '''Get the security manifest for app'''
        if app not in self.manifest['hooks']:
            error("Could not find '%s' in click manifest" % app)
        elif 'apparmor' not in self.manifest['hooks'][app]:
            error("Could not find apparmor hook for '%s' in click manifest" %
                  app)
        f = self.manifest['hooks'][app]['apparmor']
        m = self.security_manifests[f]

        return (f, m)

    def _extract_security_profile(self, app):
        '''Extract security profile'''
        rel_fn = self.manifest['hooks'][app]['apparmor-profile']

        fn = os.path.join(self.unpack_dir, rel_fn)
        if not os.path.exists(fn):
            error("Could not find '%s'" % rel_fn)

        fh = open_file_read(fn)
        contents = ""
        for line in fh.readlines():
            contents += line
        fh.close()

        # We could try to run this through apparmor_parser, but that is going
        # to be system dependent (eg, a profile may reference features on a
        # new parser and fail here on the local parser)

        return contents

    def _get_security_profile(self, app):
        '''Get the security profile for app'''
        if app not in self.manifest['hooks']:
            error("Could not find '%s' in click manifest" % app)
        elif 'apparmor-profile' not in self.manifest['hooks'][app]:
            error("Could not find apparmor-profile hook for '%s' in click "
                  "manifest" % app)
        f = self.manifest['hooks'][app]['apparmor-profile']

        p = self.security_profiles[f]
        return (f, p)

    def _get_highest_policy_version(self, vendor):
        '''Determine highest policy version for the vendor'''
        if vendor not in self.aa_policy:
            error("Could not find vendor '%s'" % vendor, do_exit=False)
            return None

        return float(sorted(self.aa_policy[vendor].keys())[-1])

    def check_policy_vendor(self):
        '''Check policy_vendor'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('policy_vendor', extra=f)
            s = "OK"
            if 'policy_vendor' in m and \
               m['policy_vendor'] not in self.aa_policy:
                t = 'error'
                s = "policy_vendor '%s' not found" % m['policy_vendor']
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('policy_vendor_matches_framework',
                                     extra=f)
            s = "OK"
            if 'policy_vendor' in m:  # policy_vendor is optional
                found_major = False
                for name, data in self.major_framework_policy.items():
                    framework = self.manifest['framework']
                    # snappy compat manifest supports comma-separated list
                    # for framework
                    if self.is_snap1 and ',' in framework:
                        # For now, we know the release framework is appended.
                        # TODO: fix for multiple frameworks
                        framework = framework.split(',')[-1]
                    if not framework.startswith(name):
                        continue
                    elif 'policy_vendor' not in data:
                        # when not specified, default to 'ubuntu'
                        data['policy_vendor'] = "ubuntu"
                    found_major = True
                    if m['policy_vendor'] != data['policy_vendor']:
                        t = 'error'
                        s = '%s != %s (%s)' % (str(m['policy_vendor']),
                                               data['policy_vendor'],
                                               framework)
                if not found_major:
                    t = 'error'
                    s = "Invalid framework '%s'" % framework
            self._add_result(t, n, s)

    def check_policy_version(self):
        '''Check policy version'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            n = self._get_check_name('policy_version_exists', extra=f)
            if 'policy_version' not in m:
                self._add_result('error', n,
                                 'could not find policy_version in manifest')
                continue

            t = 'info'
            s = "OK"
            vendor = "ubuntu"
            if 'policy_vendor' in m:
                vendor = m['policy_vendor']
            version = str(m['policy_version'])
            if vendor not in self.aa_policy or \
               not self._has_policy_version(vendor, version):
                t = 'error'
                s = 'could not find policy for %s/%s' % (vendor, version)
            self._add_result(t, n, s)

            highest = self._get_highest_policy_version(vendor)
            t = 'info'
            n = self._get_check_name(
                'policy_version_is_highest',
                extra='(%s, %s)' % (str(highest), f))
            s = "OK"
            link = None
            if float(m['policy_version']) != highest:
                t = 'info'
                link = 'http://askubuntu.com/q/562116/94326'
                s = '%s != %s' % (str(m['policy_version']), str(highest))
            self._add_result(t, n, s, link)

            t = 'info'
            n = self._get_check_name('policy_version_matches_framework',
                                     extra=f)
            s = "OK"
            found_major = False
            for name, data in self.major_framework_policy.items():
                framework = self.manifest['framework']
                # snappy compat manifest supports comma-separated list
                # for framework
                if self.is_snap1 and ',' in framework:
                    # For now, we know the release framework is appended.
                    # TODO: fix for multiple frameworks
                    framework = framework.split(',')[-1]
                if not framework.startswith(name):
                    continue
                found_major = True
                if m['policy_version'] != data['policy_version']:
                    t = 'error'
                    link = 'http://askubuntu.com/q/686347'
                    s = '%s != %s (%s)' % (str(m['policy_version']),
                                           data['policy_version'],
                                           framework)
            if not found_major:
                t = 'error'
                s = "Invalid framework '%s'" % framework
            self._add_result(t, n, s, link)

    def check_template(self):
        '''Check template'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            t = 'info'
            n = self._get_check_name('template_with_policy_version', extra=f)
            s = "OK"
            if 'policy_version' not in m:
                self._add_result('error', n,
                                 'could not find policy_version in manifest')
                continue
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('template_valid', extra=f)
            s = "OK"
            link = 'http://askubuntu.com/q/671403'
            manual_review = False
            if 'template' not in m:
                # If template not specified, we just use the default
                self._add_result(t, n, 'OK (none specified)')
                continue
            elif m['template'] in self.redflag_templates:
                t = 'warn'
                s = "(NEEDS REVIEW) '%s' not allowed" % m['template']
                manual_review = True
            elif ('policy_vendor' not in m or m['policy_vendor'] == 'ubuntu') \
                    and m['template'] in self.extraneous_ubuntu_templates:
                t = 'warn'
                s = "No need to specify '%s' template" % m['template']
            self._add_result(t, n, s, link, manual_review=manual_review)

            t = 'info'
            n = self._get_check_name('template_exists', extra=f)
            s = "OK"
            vendor = "ubuntu"
            if 'policy_vendor' in m:
                vendor = m['policy_vendor']
            version = str(m['policy_version'])

            templates = self._get_templates(vendor, version)
            if len(templates) < 1:
                t = 'error'
                s = 'could not find templates'
                self._add_result(t, n, s)
                continue
            self._add_result(t, n, s)

            found = False
            if m['template'] in self._get_templates(vendor, version):
                found = True
            elif self.is_snap1:
                frameworks = []
                if 'framework' in self.pkg_yaml:
                    frameworks = [x.strip() for x in
                                  self.pkg_yaml['framework'].split(',')]
                elif 'frameworks' in self.pkg_yaml:
                    frameworks = self.pkg_yaml['frameworks']
                elif 'type' in self.pkg_yaml and \
                     self.pkg_yaml['type'] == 'framework':
                    # frameworks may reference their own policy groups
                    frameworks.append(self.pkg_yaml['name'])

                for f in frameworks:
                    if m['template'].startswith("%s_" % f):
                        found = True
                        break

            if not found:
                t = 'error'
                s = "specified unsupported template '%s'" % m['template']

            self._add_result(t, n, s)

    def check_policy_groups_webapps(self):
        '''Check policy_groups for webapps'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('policy_groups_webapp', extra=f)
            s = "OK"
            webapp_template = "ubuntu-webapp"
            if 'template' not in m or m['template'] != webapp_template:
                # self._add_result(t, n, s)
                continue
            if 'policy_groups' not in m or \
               'networking' not in m['policy_groups']:
                self._add_result('error', n,
                                 "required group 'networking' not found")
                continue
            bad = []
            for p in m['policy_groups']:
                if p not in self.allowed_webapp_policy_groups:
                    bad.append(p)
            if len(bad) > 0:
                t = 'error'
                s = "found unusual policy groups: %s" % ", ".join(bad)
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('policy_groups_webapp_webview', extra=f)
            s = "OK"
            if self.manifest['framework'] == "ubuntu-sdk-13.10":
                s = "SKIPPED (webview not available in 13.10)"
            elif 'webview' not in m['policy_groups']:
                t = 'warn'
                s = "'webview' not specified. Webapp may not function"

            self._add_result(t, n, s)

    def check_policy_groups_push_helpers(self):
        '''Check policy groups for push-helpers'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('policy_groups_push_helper', extra=f)
            s = "OK"

            if 'template' not in m or m['template'] != 'ubuntu-push-helper':
                continue

            if 'policy_groups' not in m or \
               'push-notification-client' not in m['policy_groups']:
                self._add_result('error', n,
                                 "required group 'push-notification-client' "
                                 "not found")
                continue
            bad = []
            for p in m['policy_groups']:
                if p not in self.allowed_push_helper_policy_groups:
                    bad.append(p)
                elif p == "networking" or p == "network-client":
                    # The above covers this, but let's be very explicit and
                    # never allow networking with push-helpers
                    bad.append(p)
            if len(bad) > 0:
                t = 'error'
                s = "found unusual policy groups: %s" % ", ".join(bad)
            self._add_result(t, n, s)

    def check_policy_groups_scopes(self):
        '''Check policy_groups for scopes'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('policy_groups_scopes', extra=f)
            s = "OK"
# jdstrand, 2014-06-05: ubuntu-scope-local-content is no longer available
#            scope_templates = ['ubuntu-scope-network',
#                               'ubuntu-scope-local-content']
            scope_templates = ['ubuntu-scope-network']
            if 'template' not in m or m['template'] not in scope_templates:
                continue

            if 'policy_groups' not in m:
                continue

            bad = []
            for p in m['policy_groups']:
                if m['template'] == 'ubuntu-scope-network':
                    # networking scopes should have extremely limited access
                    if p not in self.allowed_network_scope_policy_groups:
                        bad.append(p)
# jdstrand, 2014-06-05: ubuntu-scope-local-content is no longer available
#                elif m['template'] == 'ubuntu-scope-local-content':
#                    if p == 'networking':
#                        bad.append(p)

            if len(bad) > 0:
                t = 'error'
                s = "found inappropriate policy groups: %s" % ", ".join(bad)
            self._add_result(t, n, s)

    def check_policy_groups_ubuntu_account_plugin(self):
        '''Check policy_groups for ubuntu-account-plugin template'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            if 'template' not in m or m['template'] != 'ubuntu-account-plugin':
                continue

            t = 'info'
            n = self._get_check_name(
                'required_policy_groups_ubuntu_account_plugin', extra=f)
            s = "OK"
            if 'policy_groups' not in m:
                self._add_result('error', n,
                                 "required policy groups not found")
                continue

            missing = []
            for p in self.required_ubuntu_account_plugin_policy_groups:
                if p not in m['policy_groups']:
                    missing.append(p)
            if len(missing) > 0:
                t = 'error'
                s = "missing required policy groups: %s" % ", ".join(missing)
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name(
                'policy_groups_ubuntu_account_plugin', extra=f)
            s = "OK"
            bad = []
            for p in m['policy_groups']:
                if p not in self.allowed_ubuntu_account_plugin_policy_groups:
                    bad.append(p)
            if len(bad) > 0:
                t = 'error'
                s = "found unusual policy groups: %s" % ", ".join(bad)
            self._add_result(t, n, s)

    def check_policy_groups(self):
        '''Check policy_groups'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            t = 'info'
            n = self._get_check_name('policy_groups_exists', app=app, extra=f)
            if 'policy_groups' not in m:
                # If template not specified, we just use the default
                self._add_result('info', n, 'no policy groups specified')
                continue
            elif 'policy_version' not in m:
                self._add_result('error', n,
                                 'could not find policy_version in manifest')
                continue

            s = "OK"
            vendor = "ubuntu"
            if 'policy_vendor' in m:
                vendor = m['policy_vendor']
            version = str(m['policy_version'])

            policy_groups = self._get_policy_groups(version=version,
                                                    vendor=vendor)
            if len(policy_groups) < 1:
                t = 'error'
                s = 'could not find policy groups'
                self._add_result(t, n, s)
                continue
            self._add_result(t, n, s)

            # Check for duplicates
            t = 'info'
            n = self._get_check_name('policy_groups_duplicates',
                                     app=app, extra=f)
            s = 'OK'
            tmp = []
            for p in m['policy_groups']:
                if m['policy_groups'].count(p) > 1 and p not in tmp:
                    tmp.append(p)
                if len(tmp) > 0:
                    tmp.sort()
                    t = 'error'
                    s = 'duplicate policy groups found: %s' % ", ".join(tmp)
            self._add_result(t, n, s)

            frameworks = []
            if self.is_snap1:
                if 'framework' in self.pkg_yaml:
                    frameworks = [x.strip() for x in
                                  self.pkg_yaml['framework'].split(',')]
                elif 'frameworks' in self.pkg_yaml:
                    frameworks = self.pkg_yaml['frameworks']
                elif 'type' in self.pkg_yaml and \
                     self.pkg_yaml['type'] == 'framework':
                    # frameworks may reference their own policy groups
                    frameworks.append(self.pkg_yaml['name'])

            # If we got here, we can see if valid policy groups were specified
            for i in m['policy_groups']:
                t = 'info'
                n = self._get_check_name('policy_groups_valid',
                                         app=app, extra=i)
                s = 'OK'

                # SDK will leave and empty policy group, report but don't
                # deny
                if i == "":
                    t = 'error'
                    s = 'found empty policy group'
                    self._add_result(t, n, s)
                    continue

                found = False
                framework_found = False
                for j in policy_groups:
                    if i == os.path.basename(j):
                        found = True
                        break
                    else:
                        for f in frameworks:
                            if i.startswith("%s_" % f):
                                framework_found = True
                                break
                        if framework_found:
                            found = True
                            break

                if not found:
                    t = 'error'
                    s = "unsupported policy_group '%s'" % i
                self._add_result(t, n, s)

                if found:
                    t = 'info'
                    n = self._get_check_name(
                        'policy_groups_safe', app=app, extra=i)
                    s = 'OK'
                    link = None
                    manual_review = False

                    if framework_found:
                        aa_type = 'framework'
                    else:
                        aa_type = self._get_policy_group_type(vendor, version,
                                                              i)
                    if i == "debug":
                        # IMPORTANT: do not change this string without
                        # coordination from the SDK team (LP: #1415487)
                        t = 'error'
                        s = "(REJECT) %s policy group " % aa_type + \
                            "'%s': not for production use" % (i)
                    elif aa_type == "reserved":
                        t = 'error'
                        s = "(NEEDS REVIEW) %s policy group " % aa_type + \
                            "'%s': vetted applications only" % (i)
                        if i == "debug":
                            link = 'http://askubuntu.com/a/562123/94326'
                        manual_review = True
                    elif aa_type == 'framework':
                        s = "OK (matches '%s' framework)" % i.split('_')[0]
                    elif aa_type != "common":
                        t = 'error'
                        s = "policy group '%s' has " % i + \
                            "unknown type '%s'" % (aa_type)
                    self._add_result(t, n, s, link,
                                     manual_review=manual_review)

    def check_ignored(self):
        '''Check ignored fields'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            t = 'info'
            n = self._get_check_name('ignored_fields', extra=f)
            s = "OK"
            found = []
            for i in self.ignored_fields:
                if i in m:
                    found.append(i)

            if len(found) > 0:
                t = 'warn'
                s = "found ignored fields: %s" % ", ".join(found)
            self._add_result(t, n, s)

    def check_redflag(self):
        '''Check redflag fields'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            t = 'info'
            n = self._get_check_name('redflag_fields', extra=f)
            s = "OK"
            found = []
            for i in self.redflag_fields:
                if i in m:
                    if i == 'policy_vendor' and \
                       m[i] in ['ubuntu', 'ubuntu-snappy']:
                        continue
                    found.append(i)

            if len(found) > 0:
                t = 'error'
                s = "found redflagged fields (needs human review): %s" % \
                    ", ".join(found)
            self._add_result(t, n, s)

    def check_required(self):
        '''Check required fields'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)

            t = 'info'
            n = self._get_check_name('required_fields', extra=f)
            s = "OK"
            not_found = []
            for i in self.required_fields:
                if i not in m:
                    not_found.append(i)

            if len(not_found) > 0:
                t = 'error'
                s = "missing required fields: %s" % ", ".join(not_found)
            self._add_result(t, n, s)

    def check_apparmor_profile(self):
        '''Check apparmor-profile'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps_profiles):
            (f, p) = self._get_security_profile(app)

            searches = ['###VAR###',
                        '###PROFILEATTACH###',
                        '@{CLICK_DIR}',
                        '@{APP_PKGNAME}',
                        '@{APP_VERSION}',
                        ]

            for v in searches:
                t = 'info'
                n = self._get_check_name(
                    'apparmor_profile', extra='%s (%s)' % (v, f))
                s = "OK"
                if v not in p:
                    if v.startswith('@') and \
                            ("# Unrestricted AppArmor policy" in p or
                             "# This profile offers no protection" in p):
                        self._add_result('info', n,
                                         "SKIPPED for '%s' (boilerplate)" % v)
                    else:
                        self._add_result('warn', n,
                                         "could not find '%s' in profile" % v)
                    continue
                self._add_result(t, n, s)

    # This will be really nice to get rid of when the click compat manifest
    # is gone
    def _compare_security_yamls(self, yaml, click_m):
        '''Compare two security yamls'''
        def find_match(name, key, value, my_dict):
            if 'name' in my_dict and my_dict['name'] == name and \
               key in my_dict and my_dict[key] == value:
                return True
            return False

        for first in [yaml, click_m]:
            if first == yaml:
                second = click_m
                second_m = "click-manifest"
                first_m = "package.yaml"
            else:
                second = yaml
                first_m = "click-manifest"
                second_m = "package.yaml"

            for exe_t in ['binaries', 'services']:
                t = 'info'
                n = self._get_check_name('yaml_%s' % exe_t)
                s = 'OK'
                if exe_t in first and exe_t not in second:
                    t = 'error'
                    s = "%s missing '%s'" % (second_m, exe_t)
                elif exe_t not in first and exe_t in second:
                    t = 'error'
                    s = "%s has extra '%s'" % (second_m, exe_t)
                self._add_result(t, n, s)

                if t == 'error':
                    continue
                elif exe_t not in first and exe_t not in second:
                    continue

                t = 'info'
                n = self._get_check_name('yaml_%s_entries' % exe_t)
                s = 'OK'
                if len(first[exe_t]) < len(second[exe_t]):
                    t = 'error'
                    s = "%s has extra '%s' entries" % (second_m, exe_t)
                self._add_result(t, n, s)

                for fapp in first[exe_t]:
                    t = 'info'
                    n = self._get_check_name(
                        'yaml_%s' % exe_t, app=fapp['name'])
                    s = 'OK'
                    sapp = None
                    for tmp in second[exe_t]:
                        if tmp['name'] == fapp['name']:
                            sapp = tmp
                    if sapp is None:
                        t = 'error'
                        s = "%s missing '%s'" % (second_m, fapp['name'])
                        self._add_result(t, n, s)
                        continue
                    elif first == yaml and "security-override" in fapp or \
                            second == yaml and "security-override" in sapp:
                        # no reason to check security-override since apparmor
                        # hook entry will point to this file
                        continue
                    elif first == yaml and "security-policy" in fapp or \
                            second == yaml and "security-policy" in sapp:
                        # no reason to check security-policy since apparmor
                        # profile hook is used instead
                        continue
                    elif 'caps' not in fapp and 'caps' not in sapp and \
                            second == yaml and 'security-template' not in sapp:
                        # no caps in yaml, policy_groups is empty in click
                        # manifest, unless security-template is in yaml
                        t = 'error'
                        s = "'policy_groups' not found in click manifest " + \
                            "(should default to ['network-client'])"
                        self._add_result(t, n, s)
                        continue
                    self._add_result(t, n, s)

                    for key in ['security-template', 'caps']:
                        t = 'info'
                        n = self._get_check_name(
                            'yaml_%s' % exe_t, extra=second_m)
                        s = 'OK'

                        if key not in fapp:
                            continue

                        if key == 'caps':
                            fapp['caps'] = set(fapp['caps'])
                            if 'caps' in sapp:
                                sapp['caps'] = set(sapp['caps'])

                        if not find_match(fapp['name'], key, fapp[key], sapp):
                            # handle snappy defaults for security-template
                            # and caps
                            if key == 'security-template' and \
                               second == yaml and key not in sapp and \
                               key in fapp and fapp[key] == 'default':
                                # if yaml missing security-template, click may
                                # specify 'default'
                                self._add_result(t, n, s)
                                continue
                            elif key == 'caps' and second == yaml and \
                                    'security-template' in sapp and \
                                    key not in sapp and \
                                    (key not in fapp or key in fapp and
                                     fapp[key] == set([])):
                                # when security-template is specified, then
                                # caps won't default to 'networking' when
                                # missing, so click manifest can omit or be []
                                self._add_result(t, n, s)
                                continue
                            elif key == 'caps' and second == yaml and \
                                    'security-template' not in sapp and \
                                    key not in sapp and key in fapp and \
                                    (fapp[key] == set(['networking']) or
                                     fapp[key] == set(['network-client'])):
                                # no caps in yaml, policy_groups is networking
                                # in click manifest when security-template not
                                # specified in yaml
                                self._add_result(t, n, s)
                                continue
                            elif key == 'caps' and second == click_m and \
                                    key not in sapp and key in fapp and \
                                    fapp[key] == set([]):
                                # no caps in click manifest, caps is [] in yaml
                                self._add_result(t, n, s)
                                continue
                            t = 'error'
                            s = "%s has different '%s' for '%s'" % \
                                (second_m, key, fapp['name']) + \
                                " - '%s:%s' vs '%s:%s'" % (first_m, fapp,
                                                           second_m, sapp)
                        self._add_result(t, n, s)

    def _convert_click_security_to_yaml(self):
        '''Convert click manifest to yaml'''
        converted = dict()

        # The click compat manifest does not have a 1 to 1 mapping to
        # package.yaml (ie, can't differentiate between services and
        # binaries) so look the appname up in the package.yaml to know
        # if it is a service or binary
        yaml_services = []
        if 'services' in self.pkg_yaml:
            yaml_services = [e['name'] for e in self.pkg_yaml['services']
                             if 'name' in e]
        yaml_binaries = []
        if 'binaries' in self.pkg_yaml:
            for e in self.pkg_yaml['binaries']:
                if 'exec' in e:
                    yaml_binaries.append(e['name'])
                else:
                    yaml_binaries.append(e['name'].split('/')[-1])

        for app in sorted(self.security_apps):
            if app == 'snappy-config':
                # the apparmor policy for snappy-config is autogenerated and
                # not represented in the yaml
                continue

            if app in yaml_services:
                key = 'services'
            elif app in yaml_binaries:
                key = 'binaries'
            else:
                t = 'error'
                n = self._get_check_name('yaml_and_click', app=app)
                s = "'%s' in click manifest missing from package.yaml" % app
                self._add_result(t, n, s)
                continue

            if key not in converted:
                converted[key] = []
            tmp = dict()
            tmp['name'] = app

            (f, m) = self._get_security_manifest(app)
            if 'template' in m:
                tmp['security-template'] = m['template']

            if 'policy_groups' in m:
                tmp['caps'] = m['policy_groups']

            converted[key].append(copy.deepcopy(tmp))

        for app in sorted(self.security_apps_profiles):
            if app in yaml_services:
                key = 'services'
            elif app in yaml_binaries:
                key = 'binaries'
            else:
                t = 'error'
                n = self._get_check_name('yaml_and_click', app=app)
                s = "'%s' in click manifest missing from package.yaml" % app
                self._add_result(t, n, s)
                continue

            if key not in converted:
                converted[key] = []
            tmp = dict()
            tmp['name'] = app

            (f, p) = self._get_security_profile(app)
            tmp['security-policy'] = {'apparmor': f}

            converted[key].append(copy.deepcopy(tmp))

        return copy.deepcopy(converted)

    def check_security_yaml_and_click(self):
        '''Verify click and security yaml are in sync (not including
           override)
        '''
        if not self.is_click and not self.is_snap1:
            return

        if self.is_click or self.pkg_yaml['type'] in self.sec_skipped_types:
            return

        # setup a small dict that is a subset of self.pkg_yaml
        yml = dict()
        for exe_t in ['binaries', 'services']:
            if exe_t not in self.pkg_yaml:
                continue
            yml[exe_t] = copy.deepcopy(self.pkg_yaml[exe_t])
            for item in yml[exe_t]:
                # account for binaries doing different things with 'name/exec'
                if exe_t == 'binaries' and 'exec' not in item and \
                   '/' in item['name']:
                    item['name'] = os.path.basename(item['name'])

        converted = self._convert_click_security_to_yaml()

        # don't compare the security yaml and the click if the yaml isn't
        # formatted right. This avoids confusing errors for the user
        error = False
        for exe_t in yml.keys():
            for item in yml[exe_t]:
                if 'security-template' in item and \
                        not isinstance(item['security-template'], str):
                    error = True
                    continue
                elif 'caps' in item and not isinstance(item['caps'], list):
                    error = True
                    continue
        if error:
            t = 'info'
            n = self._get_check_name('yaml_and_click')
            s = "SKIPPED (yaml errors)"
            self._add_result(t, n, s)
            return

        self._compare_security_yamls(yml, converted)

    def check_security_yaml_override_and_click(self):
        '''Verify click and security yaml override are in sync'''
        if not self.is_click and not self.is_snap1:
            return

        if self.is_click or self.pkg_yaml['type'] in self.sec_skipped_types:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue

            for item in self.pkg_yaml[exe_t]:
                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_override_click_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                app = item['name']
                t = 'info'
                n = self._get_check_name('yaml_override_click', app=app)
                s = "OK"
                if 'security-override' not in item:
                    s = "OK (skipping unspecified override)"
                elif 'apparmor' not in item['security-override']:
                    t = 'error'
                    s = "'apparmor' not specified in 'security-override' " + \
                        "for '%s'" % app
                elif item['security-override']['apparmor'] not in \
                        self.security_manifests:
                    t = 'error'
                    s = "'%s' not found in click manifest for '%s'" % \
                        (item['security-override']['apparmor'], app)
                # NOTE: we skip 'seccomp' because there isn't currently a
                # click hook for it
                self._add_result(t, n, s)

    def check_security_yaml_override(self):
        '''Verify security yaml override'''
        if not self.is_click and not self.is_snap1:
            return

        if not self.is_snap1:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue

            for item in self.pkg_yaml[exe_t]:
                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_override_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                app = item['name']
                t = 'info'
                n = self._get_check_name('yaml_override_format', app=app)
                s = "OK"
                if 'security-override' not in item:
                    s = "OK (skipping unspecified override)"
                else:
                    if 'apparmor' not in item['security-override']:
                        t = 'error'
                        s = ("'apparmor' not specified in 'security-override' "
                             "for '%s'" % app)
                    elif 'seccomp' not in item['security-override']:
                        t = 'error'
                        s = ("'seccomp' not specified in 'security-override' "
                             "for '%s'" % app)

                self._add_result(t, n, s)

    def check_security_yaml_policy(self):
        '''Verify security yaml policy'''
        if not self.is_click and not self.is_snap1:
            return

        if not self.is_snap1:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue

            for item in self.pkg_yaml[exe_t]:
                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_policy_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                app = item['name']
                t = 'info'
                n = self._get_check_name('yaml_policy_format', app=app)
                s = "OK"
                if 'security-policy' not in item:
                    s = "OK (skipping unspecified policy)"
                elif 'apparmor' not in item['security-policy']:
                    t = 'error'
                    s = "'apparmor' not specified in 'security-policy' " + \
                        "for '%s'" % app
                elif 'seccomp' not in item['security-policy']:
                    t = 'error'
                    s = "'seccomp' not specified in 'security-policy' for " + \
                        "'%s'" % app
                self._add_result(t, n, s)

                if 'security-policy' in item:
                    t = 'error'
                    n = self._get_check_name('yaml_policy_present')
                    s = "(NEEDS REVIEW) 'security-policy' not allowed"
                    link = ('https://developer.ubuntu.com/en/snappy/guides/'
                            'security-policy/')
                    m = True
                    self._add_result(t, n, s, link=link, manual_review=m)

    def check_security_yaml_combinations(self):
        '''Verify security yaml uses valid combinations'''
        if not self.is_click and not self.is_snap1:
            return

        if not self.is_snap1 or \
                self.pkg_yaml['type'] in self.sec_skipped_types:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue
            for item in self.pkg_yaml[exe_t]:
                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_combinations_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                app = item['name']

                t = 'info'
                n = self._get_check_name('yaml_combinations', app=app)
                s = "OK"
                if "security-policy" in item:
                    for i in ['security-override', 'security-template',
                              'caps']:
                        if i in item:
                            t = 'error'
                            s = "Found '%s' with 'security-policy'" % (i)
                            break
                elif "security-override" in item:
                    for i in ['security-policy', 'security-template', 'caps']:
                        if i in item:
                            t = 'error'
                            s = "Found '%s' with 'security-override'" % (i)
                            break
                self._add_result(t, n, s)

    def check_security_template(self):
        '''Check snap security-template'''
        if not self.is_click and not self.is_snap1:
            return

        if not self.is_snap1 or \
                self.pkg_yaml['type'] in self.sec_skipped_types:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue
            for item in self.pkg_yaml[exe_t]:
                if 'security-template' not in item:
                    tmpl = ""
                else:
                    tmpl = item['security-template']

                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_security-template_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                # Handle bin/exec concept with bianries
                app = os.path.basename(item['name'])

                t = 'info'
                n = self._get_check_name('yaml_security-template', app=app)
                s = "OK"
                if not isinstance(tmpl, str):
                    t = 'error'
                    s = "'%s/%s' malformed: '%s' is not str" % (exe_t, app,
                                                                tmpl)
                    self._add_result(t, n, s)
                    continue
                self._add_result(t, n, s)

                if self.is_snap1:
                    t = 'info'
                    n = self._get_check_name('yaml_security-'
                                             'template_in_manifest', app=app)
                    s = "OK"
                    if app not in self.manifest['hooks']:
                        t = 'error'
                        s = "'%s' not found in click manifest" % app
                        self._add_result(t, n, s)
                        continue
                    elif 'apparmor' not in self.manifest['hooks'][app] and \
                         'apparmor-profile' not in self.manifest['hooks'][app]:
                        t = 'error'
                        s = ("'apparmor' not found in click manifest "
                             "for '%s'" % app)
                        self._add_result(t, n, s)
                        continue

        # TODO: error if not 'common' or is 'unconfined'

    def check_security_caps(self):
        '''Check snap caps'''
        if not self.is_click and not self.is_snap1:
            return

        if not self.is_snap1 or \
                self.pkg_yaml['type'] in self.sec_skipped_types:
            return

        for exe_t in ['services', 'binaries']:
            if exe_t not in self.pkg_yaml:
                continue
            for item in self.pkg_yaml[exe_t]:
                if 'caps' not in item:
                    tmpl = []
                else:
                    tmpl = item['caps']

                if 'name' not in item:
                    t = 'error'
                    n = self._get_check_name('yaml_caps_name')
                    s = "package.yaml malformed. Could not find 'name' " + \
                        "for entry in '%s'" % item
                    self._add_result(t, n, s)
                    continue

                # Handle bin/exec concept with bianries
                app = os.path.basename(item['name'])

                t = 'info'
                n = self._get_check_name('yaml_caps', app=app)
                s = "OK"
                if not isinstance(tmpl, list):
                    t = 'error'
                    s = "'%s/%s' malformed: '%s' is not list" % (exe_t, app,
                                                                 tmpl)
                    self._add_result(t, n, s)
                    continue
                self._add_result(t, n, s)

                if self.is_snap1:
                    t = 'info'
                    n = self._get_check_name('yaml_caps_in_manifest', app=app)
                    s = "OK"
                    if app not in self.manifest['hooks']:
                        t = 'error'
                        s = "'%s' not found in click manifest" % app
                        self._add_result(t, n, s)
                        continue
                    elif 'apparmor' not in self.manifest['hooks'][app] and \
                         'apparmor-profile' not in self.manifest['hooks'][app]:
                        t = 'error'
                        s = ("'apparmor' not found in click manifest "
                             "for '%s'" % app)
                        self._add_result(t, n, s)
                        continue

        # TODO: error if not 'common'

    def check_template_online_accounts_provider(self):
        '''Check template for online accounts account-provider'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('template_account_provider', extra=f)
            s = "OK"
            if 'account-provider' not in self.manifest['hooks'][app]:
                continue
            if 'template' not in m or m['template'] != "ubuntu-account-plugin":
                t = 'error'
                s = "template is not 'ubuntu-account-plugin'"
            self._add_result(t, n, s)

    def check_template_online_accounts_qml_plugin(self):
        '''Check template for online accounts account-qml-plugin'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('template_account_qml_plugin', extra=f)
            s = "OK"
            if 'account-qml-plugin' not in self.manifest['hooks'][app]:
                continue
            if 'template' not in m or m['template'] != "ubuntu-account-plugin":
                t = 'error'
                s = "template is not 'ubuntu-account-plugin'"
            self._add_result(t, n, s)

    def check_apparmor_profile_name_length(self):
        '''Check AppArmor profile name length'''
        if not self.is_click and not self.is_snap1:
            return

        maxlen = AA_PROFILE_NAME_MAXLEN
        advlen = AA_PROFILE_NAME_ADVLEN

        for app in sorted(self.security_apps):
            (f, m) = self._get_security_manifest(app)
            t = 'info'
            n = self._get_check_name('profile_name_length', extra=f)
            s = "OK"
            profile = "%s_%s_%s" % (self.click_pkgname, app,
                                    self.click_version)
            if len(profile) > maxlen:
                t = 'error'
                s = ("'%s' too long (exceeds %d characters). Please shorten "
                     "'%s', '%s' and/or '%s'" % (profile, maxlen,
                                                 self.click_pkgname, app,
                                                 self.click_version))
            elif len(profile) > advlen:
                t = 'warn'
                s = ("'%s' is long (exceeds %d characters) and thus could be "
                     "problematic in certain environments. Please consider "
                     "shortening '%s', '%s' and/or '%s'" % (profile, advlen,
                                                            self.click_pkgname,
                                                            app,
                                                            self.click_version)
                     )
            self._add_result(t, n, s)



#
#  Copyright (C) 2014 Canonical Ltd.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

FRAMEWORKS = {
    "ubuntu-sdk-13.10": "deprecated",
    "ubuntu-sdk-14.04": "deprecated",
    "ubuntu-sdk-14.04-dev1": "obsolete",
    "ubuntu-sdk-14.04-html-dev1": "obsolete",
    "ubuntu-sdk-14.04-html": "deprecated",
    "ubuntu-sdk-14.04-papi-dev1": "obsolete",
    "ubuntu-sdk-14.04-papi": "deprecated",
    "ubuntu-sdk-14.04-qml-dev1": "obsolete",
    "ubuntu-sdk-14.04-qml": "deprecated",
    "ubuntu-sdk-14.10": "deprecated",
    "ubuntu-sdk-14.10-dev1": "obsolete",
    "ubuntu-sdk-14.10-dev2": "obsolete",
    "ubuntu-sdk-14.10-dev3": "deprecated",
    "ubuntu-sdk-14.10-html-dev1": "obsolete",
    "ubuntu-sdk-14.10-html-dev2": "obsolete",
    "ubuntu-sdk-14.10-html-dev3": "deprecated",
    "ubuntu-sdk-14.10-html": "deprecated",
    "ubuntu-sdk-14.10-papi-dev1": "obsolete",
    "ubuntu-sdk-14.10-papi-dev2": "obsolete",
    "ubuntu-sdk-14.10-papi-dev3": "deprecated",
    "ubuntu-sdk-14.10-papi": "deprecated",
    "ubuntu-sdk-14.10-qml-dev1": "obsolete",
    "ubuntu-sdk-14.10-qml-dev2": "obsolete",
    "ubuntu-sdk-14.10-qml-dev3": "deprecated",
    "ubuntu-sdk-14.10-qml": "deprecated",
    "ubuntu-sdk-15.04": "available",
    "ubuntu-sdk-15.04-html": "available",
    "ubuntu-sdk-15.04-papi": "available",
    "ubuntu-sdk-15.04-qml": "available",
    "ubuntu-sdk-15.04.1-html": "available",
    "ubuntu-sdk-15.04.1-papi": "available",
    "ubuntu-sdk-15.04.1-qml": "available",
    "ubuntu-sdk-15.04.1": "available",
    "ubuntu-sdk-15.04.2-html": "available",
    "ubuntu-sdk-15.04.2-papi": "available",
    "ubuntu-sdk-15.04.2-qml": "available",
    "ubuntu-sdk-15.04.2": "available",
    "ubuntu-sdk-15.04.3-html": "available",
    "ubuntu-sdk-15.04.3-papi": "available",
    "ubuntu-sdk-15.04.3-qml": "available",
    "ubuntu-sdk-15.04.3": "available",
    "ubuntu-sdk-15.04.4-html": "available",
    "ubuntu-sdk-15.04.4-papi": "available",
    "ubuntu-sdk-15.04.4-qml": "available",
    "ubuntu-sdk-15.04.4": "available",
    "ubuntu-sdk-15.04.5-html": "available",
    "ubuntu-sdk-15.04.5-papi": "available",
    "ubuntu-sdk-15.04.5-qml": "available",
    "ubuntu-sdk-15.04.5": "available",
    "ubuntu-sdk-15.04.6-html": "available",
    "ubuntu-sdk-15.04.6-papi": "available",
    "ubuntu-sdk-15.04.6-qml": "available",
    "ubuntu-sdk-15.04.6": "available",
    "ubuntu-sdk-15.04.7-html": "available",
    "ubuntu-sdk-15.04.7-papi": "available",
    "ubuntu-sdk-15.04.7-qml": "available",
    "ubuntu-sdk-15.04.7": "available",
    "ubuntu-sdk-16.04": "available",
    "ubuntu-sdk-16.04-html": "available",
    "ubuntu-sdk-16.04-papi": "available",
    "ubuntu-sdk-16.04-qml": "available",
    "ubuntu-sdk-16.04.1": "available",
    "ubuntu-sdk-16.04.1-html": "available",
    "ubuntu-sdk-16.04.1-papi": "available",
    "ubuntu-sdk-16.04.1-qml": "available",
    "ubuntu-sdk-16.04.2": "available",
    "ubuntu-sdk-16.04.2-html": "available",
    "ubuntu-sdk-16.04.2-papi": "available",
    "ubuntu-sdk-16.04.2-qml": "available",
    "ubuntu-sdk-16.04.3": "available",
    "ubuntu-sdk-16.04.3-html": "available",
    "ubuntu-sdk-16.04.3-papi": "available",
    "ubuntu-sdk-16.04.3-qml": "available",
    "ubuntu-sdk-16.04.4": "available",
    "ubuntu-sdk-16.04.4-html": "available",
    "ubuntu-sdk-16.04.4-papi": "available",
    "ubuntu-sdk-16.04.4-qml": "available",
    "ubuntu-sdk-16.04.5": "available",
    "ubuntu-sdk-16.04.5-html": "available",
    "ubuntu-sdk-16.04.5-papi": "available",
    "ubuntu-sdk-16.04.5-qml": "available",
    "ubuntu-sdk-16.04.6": "available",
    "ubuntu-sdk-16.04.6-html": "available",
    "ubuntu-sdk-16.04.6-papi": "available",
    "ubuntu-sdk-16.04.6-qml": "available",
    "ubuntu-sdk-16.04.7": "available",
    "ubuntu-sdk-16.04.7-html": "available",
    "ubuntu-sdk-16.04.7-papi": "available",
    "ubuntu-sdk-16.04.7-qml": "available",
    "ubuntu-sdk-16.04.8": "available",
    "ubuntu-sdk-16.04.8-html": "available",
    "ubuntu-sdk-16.04.8-papi": "available",
    "ubuntu-sdk-16.04.8-qml": "available",
    "ubuntu-sdk-20.04": "available",
    "ubuntu-sdk-20.04-html": "available",
    "ubuntu-sdk-20.04-papi": "available",
    "ubuntu-sdk-20.04-qml": "available",
}


class Frameworks(object):
    DEPRECATED_FRAMEWORKS = []
    OBSOLETE_FRAMEWORKS = []
    AVAILABLE_FRAMEWORKS = []

    def __init__(self, overrides=None):
        self.FRAMEWORKS = FRAMEWORKS
        if overrides is not None:
            self.FRAMEWORKS.update(overrides)

        for name, data in self.FRAMEWORKS.items():
            if type(data) is dict:
                state = data.get('state')
            else:
                state = data

            if state == 'deprecated':
                self.DEPRECATED_FRAMEWORKS.append(name)
            elif state == 'obsolete':
                self.OBSOLETE_FRAMEWORKS.append(name)
            elif state == 'available':
                self.AVAILABLE_FRAMEWORKS.append(name)



'''test_cr_content_hub.py: tests for the cr_content-hub module'''
#
# Copyright (C) 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from clickreviews.cr_content_hub import ClickReviewContentHub
import clickreviews.cr_tests as cr_tests


class TestClickReviewContentHub(cr_tests.TestClickReview):
    """Tests for the lint review tool."""

    def test_check_unknown_keys_none(self):
        '''Test check_unknown() - no unknown'''
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_unknown_keys()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_unknown_keys1(self):
        '''Test check_unknown() - one unknown'''
        self.set_test_content_hub(self.default_appname, "nonexistent", "foo")
        c = ClickReviewContentHub(self.test_name)
        c.check_unknown_keys()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 1, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_unknown_keys2(self):
        '''Test check_unknown() - good with one unknown'''
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        self.set_test_content_hub(self.default_appname, "nonexistent", "foo")
        c = ClickReviewContentHub(self.test_name)
        c.check_unknown_keys()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 1, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_valid_source(self):
        '''Test check_valid() - source'''
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 2, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_valid_share(self):
        '''Test check_valid() - share'''
        self.set_test_content_hub(self.default_appname, "share", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 2, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_valid_destination(self):
        '''Test check_valid() - destination'''
        self.set_test_content_hub(self.default_appname,
                                  "destination", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 2, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_valid_all(self):
        '''Test check_valid() - all'''
        self.set_test_content_hub(self.default_appname,
                                  "destination", "pictures")
        self.set_test_content_hub(self.default_appname,
                                  "share", "pictures")
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 6, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_valid_bad_value(self):
        '''Test check_valid() - bad value'''
        self.set_test_content_hub(self.default_appname, "destination", [])
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_valid_empty_value(self):
        '''Test check_valid() - empty value'''
        self.set_test_content_hub(self.default_appname, "source", "")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_peer_hooks(self):
        '''Test check_peer_hooks()'''
        self.set_test_content_hub(self.default_appname,
                                  "destination", "pictures")
        self.set_test_content_hub(self.default_appname, "share", "pictures")
        self.set_test_content_hub(self.default_appname, "source", "pictures")

        # create a new hooks database for our peer hooks tests
        tmp = dict()

        # add our hook
        tmp["content-hub"] = \
            self.test_manifest["hooks"][self.default_appname]["content-hub"]

        self.test_manifest["hooks"][self.default_appname] = tmp
        self._update_test_manifest()

        #  do the test
        c = ClickReviewContentHub(self.test_name)

        c.check_peer_hooks()
        r = c.click_report
        # We should end up with 2 info
        expected_counts = {'info': 2, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_peer_hooks_disallowed(self):
        '''Test check_peer_hooks() - disallowed'''
        self.set_test_content_hub(self.default_appname,
                                  "destination", "pictures")
        self.set_test_content_hub(self.default_appname, "share", "pictures")
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        c = ClickReviewContentHub(self.test_name)

        # create a new hooks database for our peer hooks tests
        tmp = dict()

        # add our hook
        tmp["content-hub"] = \
            self.test_manifest["hooks"][self.default_appname]["content-hub"]

        # add something not allowed
        tmp["nonexistent"] = "nonexistent-hook"

        c.manifest["hooks"][self.default_appname] = tmp
        self._update_test_manifest()

        #  do the test
        c.check_peer_hooks()
        r = c.click_report
        expected_counts = {'info': None, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_valid_snappy_1504(self):
        '''Test check_valid() - snappy 15.04'''
        self.set_test_pkgfmt("snap", "15.04")
        self.set_test_content_hub(self.default_appname,
                                  "destination", "pictures")
        self.set_test_content_hub(self.default_appname, "share", "pictures")
        self.set_test_content_hub(self.default_appname, "source", "pictures")
        c = ClickReviewContentHub(self.test_name)
        c.check_valid()
        r = c.click_report
        expected_counts = {'info': 6, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)



'''utils.py: test utils for click reviewer tools'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import shutil
import subprocess
import tempfile


def make_click(name='test', pkgfmt_type='click', pkgfmt_version='0.4',
               package_types=None, version='1.0', title="An application",
               framework='ubuntu-sdk-15.04', extra_files=None,
               output_dir=None):
    """Return the path to a click/snap package with the given data.

    Caller is responsible for deleting the output_dir afterwards.
    """
    assert(pkgfmt_type == "click" or (pkgfmt_type == "snap" and
                                      pkgfmt_version == "15.04"))

    is_snap1 = (pkgfmt_type == "snap")
    build_dir = tempfile.mkdtemp()
    package_types = package_types or []

    try:
        make_dir_structure(build_dir, pkgfmt_type=pkgfmt_type,
                           pkgfmt_version=pkgfmt_version,
                           extra_files=extra_files)
        write_icon(build_dir)

        write_manifest(build_dir, name, version,
                       title, framework, package_types,
                       is_snap1)
        write_control(build_dir, name, version, title, pkgfmt_version)
        write_preinst(build_dir)
        write_apparmor_profile(build_dir, name)
        write_other_files(build_dir)

        if pkgfmt_type == 'snap':
            write_meta_data(build_dir, name, version, title, framework)

        pkg_path = build_package(build_dir, name, version, pkgfmt_type,
                                 pkgfmt_version, output_dir=output_dir)
    finally:
        shutil.rmtree(build_dir)

    return pkg_path


def make_dir_structure(path, pkgfmt_type, pkgfmt_version, extra_files=None):
    '''Create the mandatory dir structure and extra_files. Format for
       extra_files:
         path/to/file                   create empty file in path
         path/to/dir/                   create empty dir in path
         path/to/source,path/to/link    create symlink in path
         path/to/source:path/to/link    copy source to path

         For symlink and copy, source can be an absolute path for pointing
         outside of the dir (for symlinks) or copying into the package.
    '''
    extra_files = extra_files or []

    directories = ['meta']  # write_icon() and write_manifest() assume this
    if pkgfmt_type == 'click' or pkgfmt_version == 15.04:
        directories.append('DEBIAN')

    # enumerate the directories to create
    for extra_file in extra_files:
        if ',' in extra_file:
            extra = extra_file.split(',', 1)[1]
        elif ':' in extra_file:
            extra = extra_file.split(':', 1)[1]
        else:
            extra = extra_file

        if extra.startswith('/'):
            extra = extra[1:]

        if extra.endswith('/'):
            directories.append(extra)
        else:
            directories.append(os.path.dirname(extra))

    # make the enumerated directories
    for directory in directories:
        directory = os.path.join(path, directory)
        if not os.path.exists(directory):
            os.makedirs(directory)

    for extra_file in extra_files:
        if extra_file.endswith('/'):  # nothing more to do for directories
            continue

        source_link = None
        source_path = None
        if ',' in extra_file:
            (source_link, target_path) = extra_file.split(',', 1)
        elif ':' in extra_file:
            (source_path, target_path) = extra_file.split(':', 1)
        else:
            target_path = extra_file

        dirname, basename = os.path.split(target_path)
        if basename != '':
            if source_path:
                if not source_path.startswith('/'):
                    source_path = os.path.join(path, source_path)
                shutil.copyfile(source_path, os.path.join(path, target_path))
            elif source_link:
                cur = os.getcwd()
                if target_path.startswith('/'):
                    target_path = os.path.join(path, target_path[1:])
                else:
                    os.chdir(path)
                os.symlink(source_link, target_path)
                os.chdir(cur)
            else:
                with open(os.path.join(path, target_path), 'wb'):
                    pass


def write_icon(path):
    source_path = os.path.join(os.getcwd(), 'clickreviews/data/icon.png')
    target_path = os.path.join(path, 'meta', 'icon.png')
    shutil.copyfile(source_path, target_path)


def write_manifest(path, name, version, title, framework, types, is_snap):
    manifest_content = {'framework': framework,
                        'maintainer': 'Someone <someone@example.com>',
                        'name': name,
                        'title': title,
                        'version': version,
                        'icon': 'meta/icon.png',
                        'hooks': {'app': {'apparmor':
                                          'meta/{}.apparmor'.format(name),
                                          },
                                  },
                        'description': 'This is a test app.',
                        }
    if types:
        if is_snap:
            manifest_content.update({'type': types[0]})
        else:
            if "scope" in types:
                manifest_content['hooks']['app'].update({'scope': ""})
            if "application" in types:
                manifest_content['hooks']['app'].update({'desktop': ""})

    manifest_paths = [
        os.path.join(path, 'DEBIAN', 'manifest'),
        os.path.join(path, 'manifest.json'),
    ]
    for manifest_path in manifest_paths:
        with open(manifest_path, 'w') as f:
            json.dump(manifest_content, f)


def write_meta_data(path, name, version, title, framework):
    yaml_path = os.path.join(path, 'meta', 'package.yaml')
    content = """architectures:
icon: meta/icon.png
name: {}
version: {}
framework: {}
vendor: 'Someone <someone@example.com>',
""".format(name, version, framework)

    # don't overwrite 'copy' via make_dir_structure()
    if not os.path.exists(yaml_path):
        with open(yaml_path, 'w') as f:
            f.write(content)
    with open(os.path.join(path, 'meta', 'readme.md'), 'w') as f:
        f.write(title)


def write_meta_data2(path, name, version, summary, description, yaml=None):
    yaml_path = os.path.join(path, 'meta', 'snap.yaml')
    if yaml:
        content = yaml
    else:
        content = """architectures: [ all ]
name: {}
version: {}
summary: {}
description: {}
""".format(name, version, summary, description)

    # don't overwrite 'copy' via make_dir_structure()
    if not os.path.exists(yaml_path):
        with open(yaml_path, 'w') as f:
            f.write(content)


def write_control(path, name, version, title, pkgfmt_version):
    control_path = os.path.join(path, 'DEBIAN', 'control')
    control_content = {'Package': name,
                       'Version': version,
                       'Click-Version': pkgfmt_version,
                       'Architecture': 'all',
                       'Maintainer': 'Someone <someone@example.com>',
                       'Installed-Size': '123',
                       'Description': title,
                       }
    with open(control_path, 'w') as f:
        for key, value in control_content.items():
            f.write(key + ": " + value + "\n")


def write_preinst(path):
    preinst_path = os.path.join(path, 'DEBIAN', 'preinst')
    with open(preinst_path, 'w') as f:
        f.write("""#! /bin/sh
echo "Click packages may not be installed directly using dpkg."
echo "Use 'click install' instead."
exit 1
""")
    os.chmod(preinst_path, 0o775)


def write_apparmor_profile(path, name):
    profile_path = os.path.join(path, 'meta', '{}.apparmor'.format(name))
    profile = {
        'policy_version': 1.3,
        'policy_groups': [],
    }
    with open(profile_path, 'w') as f:
        json.dump(profile, f)


def write_other_files(path):
    def write_empty_file(path, perms=0o664):
        with open(path, 'wb'):
            pass
        os.chmod(path, perms)
    write_empty_file(os.path.join(path, 'DEBIAN', 'md5sums'))


def build_package(path, name, version, pkgfmt_type, pkgfmt_version,
                  output_dir=None):
    filename = "{}_{}_all.{}".format(name, version, pkgfmt_type)
    output_dir = output_dir or tempfile.mkdtemp()
    output_path = os.path.join(output_dir, filename)

    # Note: We're not using 'click build' here as it corrects errors (such
    # as filtering out a .click directory present in the build). We want
    # to test with manually constructed, potentially tampered-with
    # clicks/snaps. Ideally, we'd be using click rather than dpkg to
    # construct the click without filtering any files in the build dir.
    subprocess.check_call(['dpkg-deb', '-b', path, output_path])

    return output_path



'''test_cr_url dispatcher.py: tests for the cr_url_dispatcher module'''
#
# Copyright (C) 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from clickreviews.cr_url_dispatcher import ClickReviewUrlDispatcher
import clickreviews.cr_tests as cr_tests


class TestClickReviewUrlDispatcher(cr_tests.TestClickReview):
    """Tests for the lint review tool."""

    def test_check_required(self):
        '''Test check_required() - has protocol'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_required_empty_value(self):
        '''Test check_required() - empty protocol'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_required_bad_value(self):
        '''Test check_required() - bad protocol'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value=[])
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_required_multiple(self):
        '''Test check_required() - multiple'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_required_multiple_with_nonexiting(self):
        '''Test check_required() - multiple with nonexistent'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com",
                                     append=True)
        self.set_test_url_dispatcher(self.default_appname,
                                     key="nonexistent",
                                     value="foo",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_optional_none(self):
        '''Test check_optional() - protocol only'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_optional_domain_suffix_empty(self):
        '''Test check_optional() - with empty domain-suffix'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_optional_domain_suffix_bad(self):
        '''Test check_optional() - with bad domain-suffix'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value=[],
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_optional_domain_suffix_nonexistent(self):
        '''Test check_optional() - with domain-suffix plus nonexistent'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com",
                                     append=True)
        self.set_test_url_dispatcher(self.default_appname,
                                     key="nonexistent",
                                     value="foo",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_optional_domain_suffix_without_protocol(self):
        '''Test check_optional() - with domain-suffix, no protocol'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_optional_domain_suffix_without_protocol2(self):
        '''Test check_optional() - with domain-suffix, nonexistent, no
           protocol'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="nonexistent",
                                     value="example.com",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_optional()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_unknown(self):
        '''Test check_unknown()'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="nonexistent",
                                     value="foo")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_unknown()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 1, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_unknown_multiple(self):
        '''Test check_unknown() - multiple with nonexistent'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="domain-suffix",
                                     value="example.com",
                                     append=True)
        self.set_test_url_dispatcher(self.default_appname,
                                     key="nonexistent",
                                     value="foo",
                                     append=True)
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_unknown()
        r = c.click_report
        expected_counts = {'info': 0, 'warn': 1, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_peer_hooks(self):
        '''Test check_peer_hooks()'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        c = ClickReviewUrlDispatcher(self.test_name)

        # create a new hooks database for our peer hooks tests
        tmp = dict()

        # add our hook
        tmp["urls"] = \
            self.test_manifest["hooks"][self.default_appname]["urls"]

        # update the manifest and test_manifest
        c.manifest["hooks"][self.default_appname] = tmp
        self._update_test_manifest()

        # do the test
        c.check_peer_hooks()
        r = c.click_report
        # We should end up with 2 info
        expected_counts = {'info': 2, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)

    def test_check_peer_hooks_disallowed(self):
        '''Test check_peer_hooks() - disallowed'''
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        c = ClickReviewUrlDispatcher(self.test_name)

        # create a new hooks database for our peer hooks tests
        tmp = dict()

        # add our hook
        tmp["urls"] = \
            self.test_manifest["hooks"][self.default_appname]["urls"]

        # add something not allowed
        tmp["nonexistent"] = "nonexistent-hook"

        c.manifest["hooks"][self.default_appname] = tmp
        self._update_test_manifest()

        # do the test
        c.check_peer_hooks()
        r = c.click_report
        expected_counts = {'info': None, 'warn': 0, 'error': 1}
        self.check_results(r, expected_counts)

    def test_check_required_snappy_1504(self):
        '''Test check_required() - has protocol - snappy 15.04'''
        self.set_test_pkgfmt("snap", "15.04")
        self.set_test_url_dispatcher(self.default_appname,
                                     key="protocol",
                                     value="some-protocol")
        c = ClickReviewUrlDispatcher(self.test_name)
        c.check_required()
        r = c.click_report
        expected_counts = {'info': 1, 'warn': 0, 'error': 0}
        self.check_results(r, expected_counts)
